
import os as _os
import json as _json

from cachetools import cached as _cached
from cachetools import LRUCache as _LRUCache

# The cache can hold a maximum of 5 objects, and will replace the least
# recently used items first
_cache_serviceinfo_data = _LRUCache(maxsize=5)
_cache_service_info = _LRUCache(maxsize=5)
_cache_adminusers = _LRUCache(maxsize=5)
_cache_serviceuser = _LRUCache(maxsize=5)
_cache_service_account_uid = _LRUCache(maxsize=5)


__all__ = ["push_is_running_service", "pop_is_running_service",
           "is_running_service", "assert_running_service",
           "setup_this_service", "add_admin_user",
           "get_this_service", "get_admin_users",
           "get_service_private_key", "save_service_keys_to_objstore",
           "get_service_private_certificate", "get_service_public_key",
           "get_service_public_certificate",
           "clear_serviceinfo_cache",
           "get_service_user_account_uid", "create_service_user_account"]


# The key in the object store for the service object
_service_key = "_service_key"

_is_running_service = 0


def push_is_running_service():
    """Internal function used to push into the 'running_service' state.
       While we are in this state then we view that all code is running
       as part of a running service
    """
    global _is_running_service
    _is_running_service += 1


def pop_is_running_service():
    """Internal function used to pop out from the 'running_service' state.
       While we are in this state then we view that all code is running
       as part of a running service
    """
    global _is_running_service
    _is_running_service -= 1

    if _is_running_service < 0:
        _is_running_service = 0


def is_running_service():
    """Return whether or not this code is running as part of a service"""
    global _is_running_service
    return _is_running_service > 0


def assert_running_service():
    """Assert that this code is running as part of a valid service"""
    if not is_running_service():
        from ._errors import ServiceError
        raise ServiceError(
            "You can only call this function from within a valid "
            "running service. A client cannot call this function.")


def clear_serviceinfo_cache():
    """Clear the caches used to accelerate loading the service info
       and admin user objects
    """
    _cache_adminusers.clear()
    _cache_service_info.clear()
    _cache_serviceinfo_data.clear()
    _cache_serviceuser.clear()
    _cache_service_account_uid.clear()


# Cache this function as the data will rarely change, and this
# will prevent too many runs to the ObjectStore
@_cached(_cache_serviceinfo_data)
def _get_this_service_data():
    """Internal function that loads up the service info data from
       the object store.
    """
    assert_running_service()

    from Acquire.Service import ServiceAccountError

    # get the bucket again - can't pass as an argument as this is a cached
    # function - luckily _get_service_account_bucket is also a cached function
    try:
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        bucket = _get_service_account_bucket()
    except ServiceAccountError as e:
        raise e
    except Exception as e:
        raise ServiceAccountError(
            "Cannot log into the service account: %s" % str(e))

    # find the service info from the object store
    try:
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        service = _ObjectStore.get_object_from_json(bucket, _service_key)
    except Exception as e:
        from Acquire.Service import MissingServiceAccountError
        raise MissingServiceAccountError(
            "Unable to load the service account for this service. An "
            "error occured while loading the data from the object "
            "store: %s" % str(e))

    if not service:
        from Acquire.Service import MissingServiceAccountError
        raise MissingServiceAccountError(
            "You haven't yet created the service account "
            "for this service. Please create an account first.")

    return service


def _get_service_password():
    """Function used to get the primary password that locks the
       skeleton key that secures the entire tree of trust from the
       Service object
    """
    service_password = _os.getenv("SERVICE_PASSWORD")

    if service_password is None:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "You must supply a $SERVICE_PASSWORD")

    return service_password


def setup_this_service(canonical_url, service_type):
    """Call this function to setup a new
       service that will serve at 'canonical_url', will be of
       the specified service_type.

        (1) Delete the object store value "_service" if you want to reset
            the actual Service. This will assign a new UID for the service
            which would reset the certificates and keys. This new service
            will need to be re-introduced to other services that need
            to trust it
    """
    assert_running_service()

    from Acquire.Service import get_service_account_bucket as \
        _get_service_account_bucket

    from Acquire.ObjectStore import Mutex as _Mutex
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.Service import Service as _Service

    bucket = _get_service_account_bucket()

    # ensure that this is the only time the service is set up
    mutex = _Mutex(key=_service_key, bucket=bucket)

    try:
        service_info = _ObjectStore.get_object_from_json(bucket, _service_key)
    except:
        service_info = None

    service = None
    service_password = _get_service_password()

    if service_info:
        try:
            service = _Service.from_data(service_info, service_password)
        except Exception as e:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "Something went write reading the Service data. You should "
                "either debug the error or delete the data at key '%s' "
                "to allow the service to be reset and constructed again. "
                "The error was %s"
                % (_service_key, str(e)))

        if service.service_type() != service_type or \
           service.canonical_url() != canonical_url:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
               "The existing service has a different type or URL to that "
               "requested at setup. The request type and URL are %s and %s, "
               "while the actual service type and URL are %s and %s." %
               (service_type, canonical_url,
                service.service_type(), service.canonical_url()))

    if service is None:
        if (service_type is None) or (canonical_url is None):
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "You need to supply both the service_type and canonical_url "
                "in order to initialise a new Service")

        # we need to build the service account
        service = _Service(service_url=canonical_url,
                           service_type=service_type)

        # now create the service user account

        # write the service data, encrypted using the service password
        service_data = service.to_data(service_password)
        # reload the data to check it is ok, and also to set the right class
        service = _Service.from_data(service_data, service_password)
        # now it is ok, save this data to the object store
        _ObjectStore.set_object_from_json(bucket, _service_key, service_data)

    mutex.unlock()

    return service


def add_admin_user(service, account_uid, authorisation=None):
    """Function that is called to add a new user account as a service
       administrator. If this is the first account then authorisation
       is not needed. If this is the second or subsequent admin account,
       then you need to provide an authorisation signed by one of the
       existing admin users. If you need to reset the admin users then
       delete the user accounts from the service.
    """
    assert_running_service()

    from Acquire.Service import get_service_account_bucket as \
        _get_service_account_bucket
    from Acquire.ObjectStore import Mutex as _Mutex
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.ObjectStore import get_datetime_now_to_string as \
        _get_datetime_now_to_string

    bucket = _get_service_account_bucket()

    # see if the admin account details exists...
    admin_key = "%s/admin_users" % _service_key

    # ensure that we have exclusive access to this service
    mutex = _Mutex(key=admin_key, bucket=bucket)

    try:
        admin_users = _ObjectStore.get_object_from_json(bucket, admin_key)
    except:
        admin_users = None

    if admin_users is None:
        # this is the first admin user - automatically accept
        admin_users = {}
        authorised_by = "first admin"
    else:
        # validate that the new user has been authorised by an existing
        # admin...
        if authorisation is None:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "You must supply a valid authorisation from an existing admin "
                "user if you want to add a new admin user.")

        if authorisation.user_uid() not in admin_users:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "The authorisation for the new admin account is not valid "
                "because the user who signed it is not a valid admin on "
                "this service.")

        authorisation.verify(account_uid)
        authorised_by = authorisation.user_uid()

    # everything is ok - add this admin user to the admin_users
    # dictionary - save the date and time they became an admin,
    # and how they achieved this status (first admin, or whoever
    # authorised them)
    admin_users[account_uid] = {"datetime": _get_datetime_now_to_string(),
                                "authorised_by": authorised_by}

    # everything is done, so now write this data to the object store
    _ObjectStore.set_object_from_json(bucket, admin_key,
                                      _json.dumps(admin_users))

    # we can (finally!) release the mutex, as everyone else should now
    # be able to see the account
    mutex.unlock()

    _cache_adminusers.clear()


@_cached(_cache_adminusers)
def get_admin_users():
    """This function returns all of the admin_users data. This is a
       dictionary of the UIDs of all of the admin users
    """
    assert_running_service()

    from Acquire.Service import ServiceAccountError

    try:
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        bucket = _get_service_account_bucket()
    except ServiceAccountError as e:
        raise e
    except Exception as e:
        raise ServiceAccountError(
            "Cannot log into the service account: %s" % str(e))

    # find the admin accounts info from the object store
    try:
        key = "%s/admin_users" % _service_key
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        admin_users = _ObjectStore.get_object_from_json(bucket, key)
    except Exception as e:
        from Acquire.Service import MissingServiceAccountError
        raise MissingServiceAccountError(
            "Unable to load the Admin User data for this service. An "
            "error occured while loading the data from the object "
            "store: %s" % str(e))

    if not admin_users:
        from Acquire.Service import MissingServiceAccountError
        raise MissingServiceAccountError(
            "You haven't yet created any Admin Users for the service account "
            "for this service. Please create an Admin User first.")

    return admin_users


@_cached(_cache_service_info)
def get_this_service(need_private_access=False):
    """Return the service info object for this service. If private
       access is needed then this will decrypt and access the private
       keys and signing certificates, which is slow if you just need
       the public certificates.
    """
    assert_running_service()

    from Acquire.Service import MissingServiceAccountError

    try:
        service_info = _get_this_service_data()
    except MissingServiceAccountError:
        raise
    except Exception as e:
        raise MissingServiceAccountError(
            "Unable to read the service info from the object store! : %s" %
            str(e))

    service_password = None

    if need_private_access:
        service_password = _get_service_password()

    try:
        from Acquire.Service import Service as _Service
        if service_password:
            service = _Service.from_data(service_info, service_password)
        else:
            service = _Service.from_data(service_info)

    except Exception as e:
        raise MissingServiceAccountError(
            "Unable to create the ServiceAccount object: %s" % str(e))

    return service


@_cached(_cache_service_account_uid)
def get_service_user_account_uid(accounting_service_uid):
    """Return the UID of the financial Acquire.Accounting.Account
       that is held on the accounting service with UID
       'accounting_service_uid' for the service user on this
       service. This is the account to which payment for this
       service should be sent
    """
    assert_running_service()

    from Acquire.Service import get_service_account_bucket as \
        _get_service_account_bucket
    from Acquire.ObjectStore import ObjectStore as _ObjectStore

    bucket = _get_service_account_bucket()

    key = "%s/account/%s" % (_service_key, accounting_service_uid)

    try:
        account_uid = _ObjectStore.get_string_object(bucket, key)
    except:
        account_uid = None

    if account_uid is None:
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "This service does not have a valid financial account on "
            "the accounting service at '%s'" % accounting_service_uid)

    return account_uid


def create_service_user_account(service, accounting_service_url):
    """Call this function to create the financial service account
       for this service on the accounting service at 'accounting_service_url'

       This does nothing if the account already exists
    """
    assert_running_service()

    accounting_service = service.get_trusted_service(accounting_service_url)
    accounting_service_uid = accounting_service.uid()

    key = "%s/account/%s" % (_service_key, accounting_service_uid)
    bucket = service.bucket()

    from Acquire.ObjectStore import ObjectStore as _ObjectStore

    try:
        account_uid = _ObjectStore.get_string_object(bucket, key)
    except:
        account_uid = None

    if account_uid:
        # we already have an account...
        return

    service_user = service.login_service_user()

    try:
        from Acquire.Client import create_account as _create_account

        account = _create_account(
                    service_user, "main",
                    "Main account to receive payment for all use on service "
                    "%s (%s)" % (service.canonical_url(), service.uid()),
                    accounting_service=accounting_service)

        account_uid = account.uid()

        _ObjectStore.set_string_object(bucket, key, account_uid)
    except Exception as e:
        from Acquire.Service import exception_to_string
        from Acquire.Service import ServiceAccountError
        raise ServiceAccountError(
            "Unable to create a financial account for the service "
            "principal for '%s' on accounting service '%s'\n\nERROR\n%s" %
            (str(service), str(accounting_service),
             exception_to_string(e)))


def _reload_key(fingerprint):
    """This function will see if we have an old key with the requested
       fingerprint, and if so, we will try to load and return that
       key from the object store
    """
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.Service import get_service_account_bucket \
        as _get_service_account_bucket
    from Acquire.Crypto import KeyManipulationError

    bucket = _get_service_account_bucket()

    try:
        key = "%s/oldkeys/fingerprints/%s" % (_service_key, fingerprint)
        keyfile = _ObjectStore.get_string_object(bucket, key)
    except:
        keyfile = None

    if keyfile is None:
        raise KeyManipulationError(
            "Cannot find a key or certificate with fingerprint '%s' : %s"
            % (fingerprint, key))

    try:
        keydata = _ObjectStore.get_object_from_json(bucket, keyfile)
    except Exception as e:
        keydata = None
        error = str(e)

    if keydata is None:
        raise KeyManipulationError(
            "Unable to load the key or certificate with fingerprint '%s': %s"
            % (fingerprint, error))

    service = get_this_service(need_private_access=True)
    return service.load_keys(keydata)[fingerprint]


def save_service_keys_to_objstore(include_old_keys=False):
    """Call this function to ensure that the current set of keys
       used for this service are saved to object store
    """
    service = get_this_service(need_private_access=True)

    oldkeys = service.dump_keys(include_old_keys=include_old_keys)

    # now write the old keys to storage
    from Acquire.ObjectStore import ObjectStore as _ObjectStore
    from Acquire.Service import get_service_account_bucket \
        as _get_service_account_bucket

    bucket = _get_service_account_bucket()

    key = "%s/oldkeys/%s" % (_service_key, oldkeys["datetime"])
    _ObjectStore.set_object_from_json(bucket, key, oldkeys)

    # now write the pointers from fingerprint to file...
    for fingerprint in oldkeys.keys():
        if fingerprint not in ["datetime", "encrypted_passphrase"]:
            _ObjectStore.set_string_object(
                bucket, "%s/oldkeys/fingerprints/%s" %
                (_service_key, fingerprint), key)


def _refresh_service_keys_and_certs(service):
    """This function will check if any key rotation is needed, and
       if so, it will automatically refresh the keys and certificates.
       The old keys and certificates will be stored in a database of
       old keys and certificates
    """
    assert_running_service()

    if not service.should_refresh_keys():
        return service

    # ensure that the current keys are saved to the object store
    save_service_keys_to_objstore()

    # generate new keys
    last_update = service.last_key_update()
    service.refresh_keys()

    # now lock the object store so that we are the only function
    # that can write the new keys to global state
    from Acquire.Service import get_service_account_bucket as \
        _get_service_account_bucket
    from Acquire.Service import Service as _Service
    from Acquire.ObjectStore import Mutex as _Mutex
    from Acquire.ObjectStore import ObjectStore as _ObjectStore

    bucket = _get_service_account_bucket()
    m = _Mutex(key=service.uid(), bucket=bucket)

    service_data = _ObjectStore.get_object_from_json(bucket, _service_key)
    service_info = _Service.from_data(service_data)

    if service_info.last_key_update() == last_update:
        # no-one else has beaten us - write the updated keys to global state
        _ObjectStore.set_object_from_json(bucket, _service_key,
                                          service.to_data(
                                              _get_service_password()))

    m.unlock()

    # clear the cache as we will need to load a new object
    clear_serviceinfo_cache()

    return get_this_service(need_private_access=True)


def get_service_private_key(fingerprint=None):
    """This function returns the private key for this service"""
    s = get_this_service(need_private_access=True)
    s = _refresh_service_keys_and_certs(s)
    key = s.private_key()

    if fingerprint:
        if key.fingerprint() != fingerprint:
            key = s.last_key()

        if key.fingerprint() != fingerprint:
            try:
                return _reload_key(fingerprint)
            except Exception as e:
                from Acquire.Service import ServiceAccountError
                raise ServiceAccountError(
                    "Cannot find a private key for '%s' that matches "
                    "the fingerprint %s. This is either because you are "
                    "using a key that is too old or "
                    "you are requesting a wrong key. Please call "
                    "refresh_keys on your Service object and try again: %s"
                    % (str(s), fingerprint, str(e)))

    return key


def get_service_private_certificate(fingerprint=None):
    """This function returns the private signing certificate
       for this service
    """
    s = get_this_service(need_private_access=True)
    s = _refresh_service_keys_and_certs(s)
    cert = s.private_certificate()

    if fingerprint:
        if cert.fingerprint() != fingerprint:
            cert = s.last_certificate()

        if cert.fingerprint() != fingerprint:
            try:
                return _reload_key(fingerprint)
            except Exception as e:
                from Acquire.Service import ServiceAccountError
                raise ServiceAccountError(
                    "Cannot find a private certificate for '%s' that matches "
                    "the fingerprint %s. This is either because you are "
                    "using a certificate that is too old or "
                    "you are requesting a wrong certificate. Please call "
                    "refresh_keys on your Service object and try again: %s"
                    % (str(s), fingerprint, str(e)))

    return cert


def get_service_public_key(fingerprint=None):
    """This function returns the public key for this service"""
    s = get_this_service(need_private_access=False)
    key = s.public_key()

    if fingerprint:
        if key.fingerprint() != fingerprint:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "Cannot find a public key for '%s' that matches "
                "the fingerprint %s" % (str(s), fingerprint))

    return key


def get_service_public_certificate(fingerprint=None):
    """This function returns the public certificate for this service"""
    s = get_this_service(need_private_access=False)
    cert = s.public_certificate()

    if fingerprint:
        if cert.fingerprint() != fingerprint:
            cert = s.last_certificate()

        if cert.fingerprint() != fingerprint:
            from Acquire.Service import ServiceAccountError
            raise ServiceAccountError(
                "Cannot find a public certificate for '%s' that matches "
                "the fingerprint %s" % (str(s), fingerprint))

    return cert




import uuid
import datetime as _datetime
import time as _time

__all__ = ["Mutex"]


class Mutex:
    """This class implements a mutex that sits in the object store.
       The mutex is associated with a key. A thread holds this mutex
       if it has successfully written its secret to this key. If
       not, then another thread must hold the mutex, and we have
       to wait...
    """
    def __init__(self, key=None, timeout=10, lease_time=10, bucket=None):
        """Create the mutex. The immediately tries to lock the mutex
           for key 'key' and will block until a lock is successfully
           obtained (or until 'timeout' seconds has been reached, and an
           exception is then thrown). If the key is provided, then
           this is the (single) global mutex. Note that this is really
           a lease, as the mutex will only be held for a maximum of
           'lease_time' seconds. After this time the mutex will be
           automatically unlocked and made available to lock by
           others. You can renew the lease by re-locking the mutex.
        """
        if key is None:
            key = "mutexes/none"
        else:
            key = "mutexes/%s" % str(key).replace(" ", "_")

        if bucket is None:
            from Acquire.Service import get_service_account_bucket as \
                                       _get_service_account_bucket

            bucket = _get_service_account_bucket()

        self._bucket = bucket
        self._key = key
        self._secret = str(uuid.uuid4())
        self._is_locked = 0
        self.lock(timeout, lease_time)

    def __del__(self):
        """Release the mutex if it is held"""
        try:
            self.fully_unlock()
        except:
            pass

    def __str__(self):
        if self.expired():
            return "Mutex(%s, EXPIRED)" % self._key
        else:
            return "Mutex(%s, is_locked=%s)" % (self._key, self.is_locked())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._key == other._key and self._secret == other._secret

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_locked(self):
        """Return whether or not this mutex is locked"""
        return self._is_locked > 0 and not self.expired()

    def seconds_remaining_on_lease(self):
        """Return the number of seconds remaining on this lease. You must
           unlock the mutex before the lease expires, or else an exception
           will be raised when you unlock, and you will likely have
           a race condition
        """
        if self.is_locked():
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            now = _get_datetime_now()

            if self._end_lease > now:
                return (self._end_lease - now).seconds
            else:
                return 0
        else:
            return 0

    def expired(self):
        """Return whether or not this lock has expired"""
        if self._is_locked > 0:
            from Acquire.ObjectStore import get_datetime_now as \
                _get_datetime_now
            return self._end_lease < _get_datetime_now()
        else:
            return False

    def assert_not_expired(self):
        """Function that asserts that this mutex has not expired"""
        if self.expired():
            from Acquire.ObjectStore import MutexTimeoutError
            raise MutexTimeoutError("The lease on this mutex expired before "
                                    "this mutex was unlocked!")

    def fully_unlock(self):
        """This fully unlocks the mutex, removing all levels
           of recursion
        """
        if self._is_locked == 0:
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now

        try:
            holder = _ObjectStore.get_string_object(self._bucket, self._key)
        except:
            holder = None

        if holder == self._lockstring:
            # we hold the mutex - delete the key
            _ObjectStore.delete_object(self._bucket, self._key)

        self._lockstring = None
        self._is_locked = 0

        if self._end_lease < _get_datetime_now():
            self._end_lease = None
            from Acquire.ObjectStore import MutexTimeoutError
            raise MutexTimeoutError("The lease on this mutex expired before "
                                    "this mutex was unlocked!")
        else:
            self._end_lease = None

    def unlock(self):
        """Release the mutex if it is held. Does nothing if the mutex
           is not held. If the mutex is unlocked after the lease has
           expired then this will raise a MutexTimeoutError. You should
           check for this when you unlock to make sure that you
           have not risked a race condition.
        """
        if self._is_locked == 0:
            return
        elif self._is_locked == 1:
            self.fully_unlock()
        else:
            self.assert_not_expired()
            self._is_locked -= 1

    def lock(self, timeout=None, lease_time=None):
        """Lock the mutex, blocking until the mutex is held, or until
           'timeout' seconds have passed. If we time out, then an exception is
           raised. The lock is held for a maximum of 'lease_time' seconds.
        """
        # if the user does not provide a timeout, then we will set a timeout
        # to 10 seconds
        if timeout is None:
            timeout = 10.0
        else:
            timeout = float(timeout)

        # if the user does not provide a lease_time, then we will set a
        # default of only 10 seconds
        if lease_time is None:
            lease_time = 10.0
        else:
            lease_time = float(lease_time)

        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        if self.is_locked():
            # renew the lease - if there is less than a second remaining
            # on the lease then unlock and then lock again from scratch
            now = _get_datetime_now()

            if (now > self._end_lease) or (now - self._end_lease).seconds < 1:
                self.fully_unlock()
                self.lock(timeout, lease_time)
            else:
                self._end_lease = now + _datetime.timedelta(seconds=lease_time)

                self._lockstring = "%s{}%s" % (
                    self._secret, _datetime_to_string(self._end_lease))

                _ObjectStore.set_string_object(self._bucket, self._key,
                                               self._lockstring)

                self._is_locked += 1

            return

        now = _get_datetime_now()
        endtime = now + _datetime.timedelta(seconds=timeout)

        # This is the first time we are trying to get a lock
        while now < endtime:
            # does anyone else hold the lock?
            try:
                holder = _ObjectStore.get_string_object(self._bucket,
                                                        self._key)
            except:
                holder = None

            is_held = True

            if holder is None:
                is_held = False
            else:
                end_lease = _string_to_datetime(holder.split("{}")[-1])
                if now > end_lease:
                    # the lease from the other holder has expired :-)
                    is_held = False

            if not is_held:
                # no-one holds this mutex - try to hold it now
                self._end_lease = now + _datetime.timedelta(seconds=lease_time)

                self._lockstring = "%s{}%s" % (
                    self._secret, _datetime_to_string(self._end_lease))

                _ObjectStore.set_string_object(self._bucket, self._key,
                                               self._lockstring)

                holder = _ObjectStore.get_string_object(self._bucket,
                                                        self._key)
            else:
                self._lockstring = None

            if holder == self._lockstring:
                # it looks like we are the holder - read and write again
                # just to make sure
                holder = _ObjectStore.get_string_object(self._bucket,
                                                        self._key)

                if holder == self._lockstring:
                    # write again just to make sure
                    _ObjectStore.set_string_object(self._bucket, self._key,
                                                   self._lockstring)

                    holder = _ObjectStore.get_string_object(self._bucket,
                                                            self._key)

            if holder == self._lockstring:
                # we have read and written our secret to the object store
                # three times. While a race condition is still possible,
                # I'd hope it is now highly unlikely - we now hold the mutex
                self._is_locked = 1
                return

            # only try the lock 4 times a second
            _time.sleep(0.25)

            now = _get_datetime_now()

        from Acquire.ObjectStore import MutexTimeoutError
        raise MutexTimeoutError("Cannot acquire a mutex lock on the "
                                "key '%s'" % self._key)




__all__ = ["DebitNote"]


class DebitNote:
    """This class holds all of the information about a completed debit. This
       is combined with credit note of equal value to form a transaction record
    """
    def __init__(self, transaction=None, account=None, authorisation=None,
                 is_provisional=False, receipt_by=None,
                 receipt=None, refund=None, bucket=None):
        """Create a debit note for the passed transaction will debit value
           from the passed account. The note will create a unique ID (uid)
           for the debit, plus the datetime of the time that value was drawn
           from the debited account. This debit note will be paired with a
           corresponding credit note from the account that received the value
           from the transaction so that a balanced TransactionRecord can be
           written to the ledger. If the note is provisional, then the value
           of the transaction will be held until the corresponding CreditNote
           has been receipted. This must be receipted before 'receipt_by',
           else the value will be returned to the DebitNote account
           (it will be automatically refunded)
        """
        self._transaction = None

        nargs = (transaction is not None) + (refund is not None) + \
                (receipt is not None)

        if nargs > 1:
            raise ValueError("You can only choose to create a debit note "
                             "from a transaction, receipt or refund!")

        if refund is not None:
            self._create_from_refund(refund, account, bucket)
        elif receipt is not None:
            self._create_from_receipt(receipt, account, bucket)
        elif (transaction is not None):
            if account is None:
                raise ValueError("You need to supply the account from "
                                 "which the transaction will be taken")

            self._create_from_transaction(transaction, account, authorisation,
                                          is_provisional, receipt_by, bucket)

    def __str__(self):
        if self.is_null():
            return "DebitNote::null"
        else:
            return "DebitNote:%s>>%s" % (self.account_uid(), self.value())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this is a null note"""
        return self._transaction is None

    def uid(self):
        """Return the UID for this note. This has the format
           dd:mm:yyyy/unique_string
        """
        if self.is_null():
            return None
        else:
            return self._uid

    def datetime(self):
        """Return the datetime for when value was debited from the account"""
        if self.is_null():
            return None
        else:
            return self._datetime

    def account_uid(self):
        """Return the UID of the account that was debited"""
        if self.is_null():
            return None
        else:
            return self._account_uid

    def transaction(self):
        """Return the transaction related to this debit note"""
        if self.is_null():
            return None
        else:
            return self._transaction

    def value(self):
        """Return the value of this note"""
        if self.is_null():
            return 0
        else:
            return self.transaction().value()

    def authorisation(self):
        """Return the authorisation that was used successfully to withdraw
           value from the debited account
        """
        if self.is_null():
            return None
        else:
            return self._authorisation

    def is_provisional(self):
        """Return whether or not the debit was provisional. Provisional debits
           are listed as liabilities
        """
        if self.is_null():
            return False
        else:
            return self._is_provisional

    def needs_receipting(self):
        """Return whether or not this DebitNote transaction needs
           receipting - if it does, then it must be receipted
           by the CreditNote before DebitNote.receipt_by().
           If this does
        """
        return self.is_provisional()

    def receipt_by(self):
        """Return the datetime by which this DebitNote must be
           receipted via the CreditNote, else the transaction
           will be automatically refunded. This will return
           'None' if the transaction has already been receipted
           or it wasn't provisional
        """
        if self.is_provisional():
            return self._receipt_by
        else:
            return None

    def _create_from_refund(self, refund, account, bucket):
        """Function used to construct a debit note by extracting
           the value specified in the passed refund from the specified
           account. This is authorised using the authorisation held in
           the refund. Note that the refund must match
           up with a prior existing provisional transaction, and this
           must not have already been refunded. This will
           actually take value out of the passed account, with that
           value residing in this debit note until it is credited to
           another account
        """
        from Acquire.Accounting import Refund as _Refund

        if not isinstance(refund, _Refund):
            raise TypeError("You can only create a DebitNote with a "
                            "Refund")

        if refund.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import TransactionState as _TransactionState
        from Acquire.Accounting import Account as _Account

        # get the transaction behind this refund and move it into
        # the "refunding" state
        transaction = _TransactionRecord.load_test_and_set(
                        refund.transaction_uid(),
                        _TransactionState.DIRECT,
                        _TransactionState.REFUNDING, bucket=bucket)

        try:
            # ensure that the receipt matches the transaction...
            transaction.assert_matching_refund(refund)

            if account is None:
                account = _Account(transaction.credit_account_uid(), bucket)
            elif account.uid() != refund.credit_account_uid():
                raise ValueError("The accounts do not match when debiting "
                                 "the refund: %s versus %s" %
                                 (account.uid(), refund.credit_account_uid()))

            # now move the refund from the credit account back to the
            # debit note
            (uid, datetime) = account._debit_refund(refund, bucket)

            self._transaction = refund.transaction()
            self._account_uid = refund.credit_account_uid()
            self._authorisation = refund.authorisation()
            self._is_provisional = False

            self._datetime = datetime
            self._uid = str(uid)
        except:
            # move the transaction back to its original state...
            _TransactionRecord.load_test_and_set(
                        refund.transaction_uid(),
                        _TransactionState.REFUNDING,
                        _TransactionState.DIRECT)
            raise

    def _create_from_receipt(self, receipt, account, bucket):
        """Function used to construct a debit note by extracting
           the value specified in the passed receipt from the specified
           account. This is authorised using the authorisation held in
           the receipt, based on the original authorisation given in the
           provisional transaction. Note that the receipt must match
           up with a prior existing provisional transaction, and this
           must not have already been receipted or refunded. This will
           actually take value out of the passed account, with that
           value residing in this debit note until it is credited to
           another account
        """
        from Acquire.Accounting import Receipt as _Receipt

        if not isinstance(receipt, _Receipt):
            raise TypeError("You can only create a DebitNote with a "
                            "Receipt")

        if receipt.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionRecord as _TransactionRecord
        from Acquire.Accounting import TransactionState as _TransactionState
        from Acquire.Accounting import Account as _Account

        # get the transaction behind this receipt and move it into
        # the "receipting" state
        transaction = _TransactionRecord.load_test_and_set(
                        receipt.transaction_uid(),
                        _TransactionState.PROVISIONAL,
                        _TransactionState.RECEIPTING, bucket=bucket)

        try:
            # ensure that the receipt matches the transaction...
            transaction.assert_matching_receipt(receipt)

            if account is None:
                account = _Account(transaction.debit_account_uid(), bucket)
            elif account.uid() != receipt.debit_account_uid():
                raise ValueError("The accounts do not match when debiting "
                                 "the receipt: %s versus %s" %
                                 (account.uid(), receipt.debit_account_uid()))

            # now move value from liability to debit, and then into this
            # debit note
            (uid, datetime) = account._debit_receipt(receipt, bucket)

            self._transaction = receipt.transaction()
            self._account_uid = receipt.debit_account_uid()
            self._authorisation = receipt.authorisation()
            self._is_provisional = False

            self._datetime = datetime
            self._uid = str(uid)
        except:
            # move the transaction back to its original state...
            _TransactionRecord.load_test_and_set(
                        receipt.transaction_uid(),
                        _TransactionState.RECEIPTING,
                        _TransactionState.PROVISIONAL)
            raise

    def _create_from_transaction(self, transaction, account, authorisation,
                                 is_provisional, receipt_by, bucket):
        """Function used to construct a debit note by extracting the
           specified transaction value from the passed account. This
           is authorised using the passed authorisation, and can be
           a provisional debit if 'is_provisional' is true. This will
           actually take value out of the passed account, with that
           value residing in this debit note until it is credited
           to another account
        """
        from Acquire.Accounting import Transaction as _Transaction
        from Acquire.Accounting import Account as _Account

        if not isinstance(transaction, _Transaction):
            raise TypeError("You can only create a DebitNote with a "
                            "Transaction")

        if not isinstance(account, _Account):
            raise TypeError("You can only create a DebitNote with a valid "
                            "Account")

        if authorisation is not None:
            from Acquire.Identity import Authorisation as _Authorisation

            if not isinstance(authorisation, _Authorisation):
                raise TypeError("Authorisation must be of type Authorisation")

        self._transaction = transaction
        self._account_uid = account.uid()
        self._authorisation = authorisation
        self._is_provisional = is_provisional

        (uid, datetime, receipt_by) = account._debit(
                                        transaction, authorisation,
                                        is_provisional, receipt_by,
                                        bucket=bucket)

        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        self._datetime = _datetime_to_datetime(datetime)
        self._uid = str(uid)

        if is_provisional:
            assert(receipt_by is not None)
            self._receipt_by = receipt_by
        else:
            assert(receipt_by is None)

    def to_data(self):
        """Return this DebitNote as a dictionary that can be encoded as json"""
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            data["transaction"] = self._transaction.to_data()
            data["account_uid"] = self._account_uid
            data["authorisation"] = self._authorisation.to_data()
            data["is_provisional"] = self._is_provisional
            data["datetime"] = _datetime_to_string(self._datetime)
            data["uid"] = self._uid

            if self._is_provisional:
                data["receipt_by"] = _datetime_to_string(self._receipt_by)

        return data

    @staticmethod
    def from_data(data):
        """Return a DebitNote that has been extracted from the passed
           json-decoded dictionary
        """
        d = DebitNote()

        if (data and len(data) > 0):
            from Acquire.Accounting import Transaction as _Transaction
            from Acquire.Identity import Authorisation as _Authorisation
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime

            d._transaction = _Transaction.from_data(data["transaction"])
            d._account_uid = data["account_uid"]
            d._authorisation = _Authorisation.from_data(data["authorisation"])
            d._is_provisional = data["is_provisional"]
            d._datetime = _string_to_datetime(data["datetime"])
            d._uid = data["uid"]

            if d._is_provisional:
                d._receipt_by = _string_to_datetime(data["receipt_by"])

        return d




__all__ = ["FileInfo", "VersionInfo"]

_version_root = "storage/version"

_fileinfo_root = "storage/file"

_file_root = "storage/file"


class VersionInfo:
    """This class holds specific info about a version of a file"""
    def __init__(self, filesize=None, checksum=None,
                 aclrules=None, compression=None, user_guid=None):
        """Construct the version of the file that has the passed
           size and checksum, was uploaded by the specified user,
           and that has the specified aclrules, and whether or not
           this file is stored and transmitted in a compressed
           state
        """
        if filesize is not None:
            from Acquire.ObjectStore import create_uuid as _create_uuid
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.Storage import ACLRules as _ACLRules

            if user_guid is None:
                raise PermissionError(
                    "You must specify the user_guid of the user who is "
                    "uploading this version of the file!")

            if aclrules is None:
                from Acquire.Storage import create_aclrules as _create_aclrules
                from Acquire.Storage import ACLRule as _ACLRule
                aclrules = _create_aclrules(aclrule=_ACLRule.inherit())
            else:
                if not isinstance(aclrules, _ACLRules):
                    raise TypeError("The aclrules must be type ACLRules")

            self._filesize = filesize
            self._checksum = checksum
            self._file_uid = _create_uuid()
            self._user_guid = str(user_guid)
            self._compression = compression
            self._aclrules = aclrules
            self._datetime = _get_datetime_now()

        else:
            self._filesize = None

    def is_null(self):
        """Return whether or not this is null"""
        return self._filesize is None

    def filesize(self):
        """Return the size in bytes of this version of the file"""
        if self.is_null():
            return 0
        else:
            return self._filesize

    def checksum(self):
        """Return the checksum for this version of the file"""
        if self.is_null():
            return None
        else:
            return self._checksum

    def aclrules(self):
        """Return all of the ACL rules for this version of the file"""
        if self.is_null():
            return None
        else:
            return self._aclrules

    def uid(self):
        """Return the UID of this version of the file in object store"""
        if self.is_null():
            return None
        else:
            return self._file_uid

    def is_compressed(self):
        """Return whether or not this file is stored and transmitted
           in a compressed state
        """
        if self.is_null():
            return False
        else:
            return self._compression is not None

    def compression_type(self):
        """Return the type of compression used if this file is
           stored and transmitted in a compressed state, or None
           if this is not compressed
        """
        if self.is_null():
            return None
        else:
            return self._compression

    def datetime(self):
        """Return the datetime when this version was created"""
        if self.is_null():
            return None
        else:
            return self._datetime

    def uploaded_by(self):
        """Return the GUID of the user that uploaded this version"""
        if self.is_null():
            return None
        else:
            return self._user_guid

    def _file_key(self):
        """Return the key for this actual file for this version
           in the object store"""
        if self.is_null():
            return None
        else:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string

            return "%s/%s/%s" % (_file_root,
                                 _datetime_to_string(self._datetime),
                                 self._file_uid)

    def _key(self, drive_uid, encoded_filename):
        """Return the key for this version in the object store"""
        if self.is_null():
            return None
        else:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            return "%s/%s/%s/%s/%s" % (
                _version_root, drive_uid, encoded_filename,
                _datetime_to_string(self._datetime), self._file_uid)

    def to_data(self):
        """Return a json-serialisable dictionary for this object"""
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            from Acquire.ObjectStore import dict_to_string \
                as _dict_to_string
            data["filesize"] = self._filesize
            data["checksum"] = self._checksum
            data["file_uid"] = self._file_uid
            data["datetime"] = _datetime_to_string(self._datetime)
            data["user_guid"] = self._user_guid

            if self._aclrules is not None:
                data["aclrules"] = self._aclrules.to_data()

            if self._compression is not None:
                data["compression"] = self._compression

        return data

    @staticmethod
    def from_data(data):
        """Return this object constructed from the passed json-deserialised
           dictionary
        """
        v = VersionInfo()

        if data is not None and len(data) > 0:
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            v._filesize = data["filesize"]
            v._checksum = data["checksum"]
            v._file_uid = data["file_uid"]
            v._user_guid = data["user_guid"]
            v._datetime = _string_to_datetime(data["datetime"])

            if "aclrules" in data:
                from Acquire.Storage import ACLRules as _ACLRules
                v._aclrules = _ACLRules.from_data(data["aclrules"])

            if "compression" in data:
                v._compression = data["compression"]
            else:
                v._compression = None

        return v


class FileInfo:
    """This class provides information about a user-file that has
       been uploaded to the storage service. This includes all
       versions of the file, the ACLs for different users etc.

       Just as Acquire.Client.Drive provides the client-side view
       of Acquire.Storage.DriveInfo, so to does
       Acquire.Client.FileHandle provide the client-side view
       of Acquire.Storage.FileInfo
    """
    def __init__(self, drive_uid=None, filehandle=None, user_guid=None):
        """Construct from a passed filehandle of a file that will be
           uploaded
        """
        self._filename = None

        if filehandle is not None:
            from Acquire.Client import FileHandle as _FileHandle

            if not isinstance(filehandle, _FileHandle):
                raise TypeError(
                    "The filehandle must be of type FileHandle")

            if filehandle.is_null():
                return

            self._drive_uid = drive_uid

            from Acquire.ObjectStore import string_to_encoded \
                as _string_to_encoded
            from Acquire.ObjectStore import string_to_filepath \
                as _string_to_filepath

            self._filename = _string_to_filepath(filehandle.filename())
            self._encoded_filename = _string_to_encoded(self._filename)

            version = VersionInfo(filesize=filehandle.filesize(),
                                  checksum=filehandle.checksum(),
                                  user_guid=user_guid,
                                  compression=filehandle.compression_type(),
                                  aclrules=filehandle.aclrules())

            self._latest_version = version

    def is_null(self):
        """Return whether or not this is null"""
        return self._filename is None

    def filename(self):
        """Return the object-store filename for this file"""
        return self._filename

    @staticmethod
    def _get_filemeta(filename, version, resolved_acl=None):
        """Internal function used to create a FileMeta from the passed
           filename and VersionInfo object
        """
        from Acquire.Client import FileMeta as _FileMeta

        filemeta = _FileMeta(filename=filename, uid=version.uid(),
                             filesize=version.filesize(),
                             checksum=version.checksum(),
                             uploaded_by=version.uploaded_by(),
                             uploaded_when=version.datetime(),
                             compression=version.compression_type(),
                             aclrules=version.aclrules())

        if resolved_acl is not None:
            filemeta.resolve_acl(resolved_acl=resolved_acl)

        return filemeta

    def get_filemeta(self, version=None, resolved_acl=None):
        """Return the metadata about the latest (or specified) version
           of this file. If 'resolved_acl' is specified, then
           return the
        """
        from Acquire.Client import FileMeta as _FileMeta

        if self.is_null():
            return _FileMeta()

        return FileInfo._get_filemeta(filename=self._filename,
                                      version=self._version_info(version),
                                      resolved_acl=resolved_acl)

    def _version_info(self, version=None):
        """Return the version info object of the latest version of
           the file, or the passed version
        """
        if self.is_null():
            return VersionInfo()
        else:
            if version is None:
                return self._latest_version

            from Acquire.Storage import MissingVersionError
            raise MissingVersionError(
                "Cannot find the version '%s' for file '%s'" %
                (version, self.filename()))

    def filesize(self, version=None):
        """Return the size (in bytes) of the latest (or specified)
           version of this file"""
        return self._version_info(version=version).filesize()

    def checksum(self, version=None):
        """Return the checksum of the latest (or specified) version
           of this file
        """
        return self._version_info(version=version).checksum()

    def is_compressed(self, version=None):
        """Return whether or not the latest (or specified) version
           of this file is stored and transmitted in a compressed
           state
        """
        return self._version_info(version=version).is_compressed()

    def compression_type(self, version=None):
        """Return the compression type (or None if not compressed)
           of the latest (or specified) version of this file
        """
        return self._version_info(version=version).compression_type()

    def drive_uid(self):
        """Return the UID of the drive on which this file resides"""
        return self._drive_uid

    def drive(self):
        """Return the actual DriveInfo object for the drive on which this
           file resides
        """
        if self.is_null():
            return None
        else:
            from Acquire.Storage import DriveInfo as _DriveInfo
            return _DriveInfo(drive_uid=self.drive_uid())

    def file_uid(self, version=None):
        """Return the UID of the latest (or specified) version
           of this file
        """
        return self._version_info(version=version).uid()

    def aclrules(self, version=None):
        """Return the ACL rules for the specified user, or if that is not
           specified, the ACL rules for the current version
        """
        return self._version_info(version=version).aclrules()

    def version(self, version):
        """Return the version at the specified datetime"""
        return self._version_info(version=version)

    def latest_version(self):
        """Return the latest version of this file on the storage service. This
           is a datetime of the upload of the latest version. You will need to
           use the 'versions' function to find if there are other versions.
        """
        if self.is_null():
            return None
        else:
            return self._latest_version

    def versions(self):
        """Return the sorted list of all versions of this file on the
           storage service
        """
        if self.is_null():
            return []
        else:
            return {self._latest_version.datetime(), self._latest_version}

    def _fileinfo_key(self):
        """Return the key for this fileinfo in the object store"""
        return "%s/%s/%s" % (_fileinfo_root, self._drive_uid,
                             self._encoded_filename)

    def save(self):
        """Save this fileinfo to the object store"""
        if self.is_null():
            return

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        metadata_bucket = self.drive()._get_metadata_bucket()

        # save the version information (saves old versions)
        _ObjectStore.set_object_from_json(
                        bucket=metadata_bucket,
                        key=self._latest_version._key(self._drive_uid,
                                                      self._encoded_filename),
                        data=self._latest_version.to_data())

        # save the fileinfo itself
        _ObjectStore.set_object_from_json(bucket=metadata_bucket,
                                          key=self._fileinfo_key(),
                                          data=self.to_data())

    @staticmethod
    def list_versions(drive, filename, include_metadata=False, **kwargs):
        """List all of the versions of this file. If 'include_metadata'
           is True then this will load all of the associated metadata
           for each file
        """
        from Acquire.Storage import DriveInfo as _DriveInfo
        from Acquire.Storage import FileMeta as _FileMeta

        if not isinstance(drive, _DriveInfo):
            raise TypeError("The drive must be of type DriveInfo")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        metadata_bucket = drive._get_metadata_bucket()

        encoded_filename = _string_to_encoded(filename)

        version_root = "%s/%s/%s/" % (
                _version_root, drive.uid(), encoded_filename)

        versions = []

        if include_metadata:
            objs = _ObjectStore.get_all_objects_from_json(
                                            bucket=metadata_bucket,
                                            prefix=version_root)

            for data in objs.values():
                version = VersionInfo.from_data(data)
                filemeta = FileInfo._get_filemeta(filename=filename,
                                                  version=version)
                versions.append(filemeta)
        else:
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            keys = _ObjectStore.get_all_object_names(
                                            bucket=metadata_bucket,
                                            prefix=version_root)

            for key in keys:
                parts = key.split("/")
                uid = parts[-1]
                uploaded_when = _string_to_datetime(parts[-2])
                filemeta = _FileMeta(filename=filename,
                                     uploaded_when=uploaded_when,
                                     uid=uid)
                versions.append(filemeta)

        return versions

    @staticmethod
    def load(drive, filename, version=None):
        """Load and return the FileInfo for the file called 'filename'
           on the passed 'drive'. """
        from Acquire.Storage import DriveInfo as _DriveInfo

        if not isinstance(drive, _DriveInfo):
            raise TypeError("The drive must be of type DriveInfo")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        metadata_bucket = drive._get_metadata_bucket()

        encoded_filename = _string_to_encoded(filename)

        filekey = "%s/%s/%s" % (_fileinfo_root, drive.uid(),
                                encoded_filename)

        try:
            data = _ObjectStore.get_object_from_json(bucket=metadata_bucket,
                                                     key=filekey)
        except Exception as e:
            print(e)
            data = None

        if data is None:
            from Acquire.Storage import MissingFileError
            raise MissingFileError(
                "Cannot find the file called '%s' on drive '%s'" %
                (filename, drive))

        f = FileInfo.from_data(data)
        f._drive_uid = drive.uid()

        if version is not None:
            f._latest_version = f._version_info(version=version)

        return f

    def to_data(self):
        """Return a json-serialisable dictionary for this object"""
        data = {}

        if not self.is_null():
            data["filename"] = self.filename()
            data["latest_version"] = self.latest_version().to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return this object constructed from the passed json-deserialised
           dictionary
        """
        f = FileInfo()

        if data is not None and len(data) > 0:
            f._filename = data["filename"]
            f._latest_version = VersionInfo.from_data(data["latest_version"])
            f._drive_uid = None

        return f



"""
Acquire : (C) Christopher Woods 2018

This module contains the interfaces to the object store that provides the
storage of all state for the system. As such, this is the foundation
module that is unlikely to be used by the user, but instead is used
by most of the other modules.
"""

from ._objstore import *
from ._encoding import *
from ._mutex import *
from ._errors import *

try:
    if __IPYTHON__:
        def _set_printer(C):
            """Function to tell ipython to use __str__ if available"""
            get_ipython().display_formatter.formatters['text/plain'].for_type(
                C,
                lambda obj, p, cycle: p.text(str(obj) if not cycle else '...')
                )

        import sys as _sys
        import inspect as _inspect

        _clsmembers = _inspect.getmembers(_sys.modules[__name__],
                                          _inspect.isclass)

        for _clsmember in _clsmembers:
            _set_printer(_clsmember[1])
except:
    pass




__all__ = ["create_new_file", "compress", "uncompress"]


def compress(inputfile=None, outputfile=None,
             inputdata=None, compression_type="bz2"):
    """Compress either the passed filename or filedata using the
       specified compression type. This will compress either to the
       file called 'outputfile', or to a tmpfile. The name of the
       file will be returned. If this is compressing data,
       then it will return the compressed data
    """
    import os as _os

    if compression_type == "bz2":
        import bz2 as _bz2

        if inputfile is not None:
            IFILE = open(inputfile, "rb")

            # compress to a tmpfile and then move to outputfile later...
            import tempfile as _tempfile
            (fd, bz2file) = _tempfile.mkstemp(dir=".")
            _os.close(fd)

            try:
                OFILE = _bz2.BZ2File(bz2file, "wb", compresslevel=9)

                # compress data in MB blocks
                data = IFILE.read(1048576)

                while data:
                    OFILE.write(data)
                    data = IFILE.read(1048576)

                IFILE.close()
                OFILE.close()
            except Exception as e:
                print(e)
                # make sure we delete the temporary bz2file
                _os.unlink(bz2file)
                raise

            if outputfile is None:
                return bz2file

            try:
                # move the bz2file to the correct output name
                _os.rename(bz2file, outputfile)
                return outputfile
            except Exception as e:
                print(e)
                # we can't rename the file - just return the bz2 name
                return bz2file

        elif inputdata is not None:
            # compress the passed data and return
            return _bz2.compress(inputdata)
    else:
        raise ValueError("Unrecognised compression type '%s'" %
                         compression_type)


def uncompress(inputfile=None, outputfile=None,
               inputdata=None, compression_type="bz2"):
    """Uncompress either the passed filename or filedata using the
       specified compression type. This will uncompress either to the
       file called 'outputfile', or to a tmpfile. The name of the
       file will be returned. If this is uncompressing data,
       then it will return the uncompressed data
    """
    import os as _os

    if compression_type == "bz2":
        import bz2 as _bz2

        if inputfile is not None:
            IFILE = _bz2.BZ2File(inputfile, "rb")

            # compress to a tmpfile and then move to outputfile later...
            import tempfile as _tempfile
            (fd, bz2file) = _tempfile.mkstemp(dir=".")
            _os.close(fd)

            try:
                OFILE = open(bz2file, "wb")

                # compress data in MB blocks
                data = IFILE.read(1048576)

                while data:
                    OFILE.write(data)
                    data = IFILE.read(1048576)

                IFILE.close()
                OFILE.close()
            except Exception as e:
                print(e)
                # make sure we delete the temporary bz2file
                _os.unlink(bz2file)
                raise

            if outputfile is None:
                return bz2file

            try:
                # move the bz2file to the correct output name
                _os.rename(bz2file, outputfile)
                return outputfile
            except Exception as e:
                print(e)
                # we can't rename the file - just return the bz2 name
                return bz2file

        elif inputdata is not None:
            # compress the passed data and return
            return _bz2.decompress(inputdata)
    else:
        raise ValueError("Unrecognised compression type '%s'" %
                         compression_type)


def create_new_file(filename, dir=None):
    """Create a new file in directory 'dir' (default current directory)
       called 'filename'. If the file already exists, then create a new
       file with name derived from 'filename'
    """
    import os as _os

    filename = _os.path.split(filename)[1]

    if filename.find(".") != -1:
        parts = filename.split(".")
        base = ".".join(parts[0:-1])
        ext = parts[-1]
        if len(ext) == 0:
            ext = None
    else:
        base = filename
        ext = None

    if dir is not None:
        filebase = _os.path.join(dir, base)
    else:
        filebase = base

    if ext is None:
        filename = filebase
    else:
        filename = "%s.%s" % (filebase, ext)

    while True:
        idx = 1
        while _os.path.exists(filename):
            filename = "%s_%s.%s" % (filebase, idx, ext)
            idx += 1

        try:
            # open in "x" move should prevent race condition
            FILE = open(filename, "x")
            FILE.close()
            return _os.path.realpath(filename)
        except:
            pass




__all__ = ["AccountError", "PackingError", "UnpackingError",
           "RemoteFunctionCallError", "ServiceError", "ServiceAccountError",
           "MissingServiceAccountError"]


class AccountError(Exception):
    pass


class PackingError(Exception):
    pass


class UnpackingError(Exception):
    pass


class RemoteFunctionCallError(Exception):
    """This exception is called if there is a remote function
       call error that could not be auto-converted to anything
       else. If a child exception occurred remotely, then this
       is packaged safely into this exception
    """
    def __init__(self, message, child_exception=None):
        super().__init__(message)

        if child_exception is not None:
            if issubclass(child_exception.__class__, Exception):
                from ._function import exception_to_safe_exception \
                    as _exception_to_safe_exception

                self._child_exception = _exception_to_safe_exception(
                                                            child_exception)
            else:
                self.args = ("%s : %s" % (message, str(child_exception)), )

    def unpack_and_raise(self):
        if self._child_exception is None:
            raise self
        else:
            import traceback as _traceback
            tb_lines = _traceback.format_exception(
                self.__class__, self, self.__traceback__)

            self._child_exception.args = (
                "%s\n=== Local Traceback ===\n%s"
                % (self._child_exception.args[0],
                   "".join(tb_lines)), )

            if isinstance(self._child_exception, RemoteFunctionCallError):
                self._child_exception.unpack_and_raise()
            else:
                raise self._child_exception


class ServiceError(Exception):
    pass


class ServiceAccountError(Exception):
    pass


class MissingServiceAccountError(Exception):
    pass





__all__ = ["ObjectStoreError", "MutexTimeoutError", "EncodingError"]


class ObjectStoreError(Exception):
    pass


class EncodingError(ObjectStoreError):
    pass


class MutexTimeoutError(Exception):
    pass




import os as _os
import base64 as _base64
import uuid as _uuid

from Acquire.Stubs import lazy_import as _lazy_import

_rsa = _lazy_import.lazy_module(
            "cryptography.hazmat.primitives.asymmetric.rsa")
_serialization = _lazy_import.lazy_module(
            "cryptography.hazmat.primitives.serialization")
_default_backend = _lazy_import.lazy_function(
            "cryptography.hazmat.backends.default_backend")
_hashes = _lazy_import.lazy_module("cryptography.hazmat.primitives.hashes")
_padding = _lazy_import.lazy_module(
            "cryptography.hazmat.primitives.asymmetric.padding")
_fernet = _lazy_import.lazy_module("cryptography.fernet")

__all__ = ["PrivateKey", "PublicKey", "get_private_key"]


def _bytes_to_string(b):
    """Return the passed binary bytes safely encoded to
       a base64 utf-8 string"""
    if b is None:
        return None
    else:
        return _base64.b64encode(b).decode("utf-8")


def _string_to_bytes(s):
    """Return the passed base64 utf-8 encoded binary data
       back converted from a string back to bytes. Note that
       this can only convert strings that were encoded using
       bytes_to_string - you cannot use this to convert
       arbitrary strings to bytes"""
    if s is None:
        return None
    else:
        return _base64.b64decode(s.encode("utf-8"))


def _assert_strong_passphrase(passphrase, mangleFunction):
    """This function returns whether or not the passed
       passphrase is sufficiently strong. To be strong,
       the password must be between 6-50 characters,
       mix upper and lower case, and contain letters and
       numbers
    """

    if mangleFunction:
        passphrase = str(mangleFunction(passphrase))
    else:
        passphrase = str(passphrase)

    if len(passphrase) < 6 or len(passphrase) > 50:
        from Acquire.Crypto import WeakPassphraseError
        raise WeakPassphraseError("The pass-phrase '%s' must contain between "
                                  "6 and 50 characters" % passphrase)

    import re as _re

    if not (_re.search(r'[A-Z]', passphrase) and
            _re.search(r'[a-z]', passphrase) and
            _re.search(r'[0-9]', passphrase)):
        from Acquire.Crypto import WeakPassphraseError
        raise WeakPassphraseError("The pass-phrase must contain numbers and "
                                  "upper- and lowercase characters")

    return passphrase


def _generate_private_key():
    """Internal function that is used to generate all of our private keys"""
    return _rsa.generate_private_key(public_exponent=65537,
                                     key_size=2048,
                                     backend=_default_backend())


_key_database = {}


def get_private_key(key="default"):
    """Internal function used to return the key associated with
       'key'. This allows a single session to re-use a key, rather
       than continually generating new keys (which is slow). Make
       sure you know that you can safely re-use a key. Re-using
       keys for different function calls is completely ok
    """
    global _key_database

    if key in _key_database:
        return _key_database[key]
    else:
        privkey = PrivateKey()
        _key_database[key] = privkey
        return privkey


class PublicKey:
    """This is a holder for an in-memory public key"""
    def __init__(self, public_key=None):
        """Construct from the passed public key"""
        self._pubkey = public_key

    def bytes(self):
        """Return the raw bytes for this key"""
        if self._pubkey is None:
            return None

        return self._pubkey.public_bytes(
            encoding=_serialization.Encoding.PEM,
            format=_serialization.PublicFormat
                                 .SubjectPublicKeyInfo)

    def pem(self):
        """Return a PEM string for this key"""
        return self.bytes().decode("utf-8")

    def __str__(self):
        """Return a string representation of this key"""
        return "PublicKey('%s')" % self.bytes().decode("utf-8")

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.bytes() == other.bytes()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def write(self, filename):
        """Write this public key to 'filename'"""
        if self._pubkey is None:
            return

        pubkey_bytes = self.bytes()

        with open(filename, "wb") as FILE:
            FILE.write(pubkey_bytes)

    @staticmethod
    def read_bytes(data):
        """Read and return a public key from 'data'"""
        public_key = _serialization.load_pem_public_key(
                        data, backend=_default_backend())

        return PublicKey(public_key)

    @staticmethod
    def read(filename):
        """Read and return a public key from 'filename'"""
        with open(filename, "rb") as FILE:
            return PublicKey.read_bytes(FILE.read())

    def fingerprint(self):
        """Return the fingerprint of this key - this is useful to help
           work out which key to use to decrypt data
        """
        from hashlib import md5 as _md5
        md5 = _md5()
        md5.update(self.bytes())
        h = md5.hexdigest()
        # return this signature as "AA:BB:CC:DD:EE:etc."
        return ":".join([h[i:i+2] for i in range(0, len(h), 2)])

    def encrypt(self, message):
        """Encrypt and return the passed message. For short messages this
           will use the private key directly. For longer messages,
           this will generate a random
           symmetric key, will encrypt the message using that, and will then
           encrypt the symmetric key. This returns some bytes
        """
        if isinstance(message, str):
            message = message.encode("utf-8")

        try:
            return self._pubkey.encrypt(
                        message,
                        _padding.OAEP(
                            mgf=_padding.MGF1(algorithm=_hashes.SHA256()),
                            algorithm=_hashes.SHA256(),
                            label=None)
                        )
        except:
            pass

        # this is a longer message that cannot be encoded using
        # an asymmetric key - need to use a symmetric key
        key = _fernet.Fernet.generate_key()
        f = _fernet.Fernet(key)
        token = f.encrypt(message)

        encrypted_key = self._pubkey.encrypt(
                            key,
                            _padding.OAEP(
                                mgf=_padding.MGF1(algorithm=_hashes.SHA256()),
                                algorithm=_hashes.SHA256(),
                                label=None)
                            )

        # the first 256 bytes are the encrypted key - the rest
        # is the token, because we are using 2048 bit (256 byte) keys
        return encrypted_key + token

    def verify(self, signature, message):
        """Verify that the message has been correctly signed"""
        if self._pubkey is None:
            from Acquire.Crypto import KeyManipulationError
            raise KeyManipulationError("You cannot verify a message using "
                                       "an empty public key!")

        if isinstance(message, str):
            message = message.encode("utf-8")

        try:
            self._pubkey.verify(
                          signature,
                          message,
                          _padding.PSS(
                             mgf=_padding.MGF1(_hashes.SHA256()),
                             salt_length=_padding.PSS.MAX_LENGTH),
                          _hashes.SHA256())
        except Exception as e:
            from Acquire.Crypto import SignatureVerificationError
            raise SignatureVerificationError(
                       "Error validating the signature "
                       "for the passed message: %s" % str(e))

    def to_data(self):
        """Return this public key as a json-serialisable dictionary"""
        data = {}

        b = self.bytes()

        if b is not None:
            data["bytes"] = _bytes_to_string(self.bytes())

        return data

    @staticmethod
    def from_data(data):
        """Construct from the passed json-deserialised dictionary"""
        if isinstance(data, str):
            return PublicKey.read_bytes(_string_to_bytes(data))

        elif isinstance(data, bytes):
            return PublicKey.read_bytes(data)

        else:
            key = PublicKey()

            if (data and len(data) > 0):
                key = PublicKey.read_bytes(_string_to_bytes(data["bytes"]))

            return key


class PrivateKey:
    """This is a holder for an in-memory private key"""
    def __init__(self, private_key=None, auto_generate=True):
        """Construct the key either from a passed key, or by generating
           a new key"""
        self._privkey = private_key

        if self._privkey is None:
            if auto_generate:
                self._privkey = _generate_private_key()

    def __str__(self):
        """Return a string representation of this key"""
        return "PrivateKey( public_key='%s' )" % \
            self.public_key().bytes().decode("utf-8")

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.public_key() == other.public_key()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def read_bytes(data, passphrase, mangleFunction=None):
        """Read a private key from the passed bytes 'data' that
           is encrypted using 'passphrase' and return a PrivateKey
           object holding that key
        """

        passphrase = _assert_strong_passphrase(passphrase, mangleFunction)

        private_key = None

        try:
            private_key = _serialization.load_pem_private_key(
                             data,
                             password=passphrase.encode("utf-8"),
                             backend=_default_backend())
        except Exception as e:
            from Acquire.Crypto import KeyManipulationError
            raise KeyManipulationError("Cannot unlock key. %s" %
                                       str(e))

        return PrivateKey(private_key)

    @staticmethod
    def read(filename, passphrase, mangleFunction=None):
        """Read a private key from 'filename' that is encrypted using
           'passphrase' and return a PrivateKey object holding that
           key"""

        data = None

        try:
            with open(filename, "rb") as FILE:
                data = FILE.read()
        except IOError as e:
            from Acquire.Crypto import KeyManipulationError
            raise KeyManipulationError(
                    "Cannot read the private keyfile %s: %s" %
                    (filename, str(e)))

        return PrivateKey.read_bytes(data, passphrase, mangleFunction)

    @staticmethod
    def random_passphrase():
        """Randomly generate and return a passphrase that obeys the
           password rules and could be used to serialise a PrivateKey
        """
        import random as _random
        import string as _string

        # use the operating system as source of random numbers
        rand = _random.SystemRandom()

        # generate a random password comprised of a random set of
        # upper, lower and digits characters
        nvals = int(rand.uniform(20, 40))
        nlower = int(rand.uniform(1, nvals-5))
        nupper = int(rand.uniform(1, nvals-nlower-5))
        ndigits = nvals - nupper - nlower

        assert(nlower > 0)
        assert(nupper > 0)
        assert(ndigits > 0)

        lower = [rand.choice(_string.ascii_lowercase)
                 for _ in range(nlower)]
        upper = [rand.choice(_string.ascii_uppercase)
                 for _ in range(nupper)]
        digits = [rand.choice(_string.digits)
                  for _ in range(ndigits)]

        passphrase = "".join(rand.sample(lower+upper+digits, nvals))
        _assert_strong_passphrase(passphrase, mangleFunction=None)
        return passphrase

    def pem(self, passphrase, mangleFunction=None):
        """Return a PEM string for this key"""
        return self.bytes(passphrase, mangleFunction).decode("utf-8")

    def bytes(self, passphrase, mangleFunction=None):
        """Return the raw bytes for this key, encoded by the passed
           passphrase that has been optionally mangled by mangleFunction"""
        if self._privkey is None:
            return None

        passphrase = _assert_strong_passphrase(passphrase, mangleFunction)

        return self._privkey.private_bytes(
                encoding=_serialization.Encoding.PEM,
                format=_serialization.PrivateFormat.PKCS8,
                encryption_algorithm=_serialization.BestAvailableEncryption(
                                        passphrase.encode("utf-8")))

    def write(self, filename, passphrase, mangleFunction=None):
        """Write this key to 'filename', encrypted with 'passphrase'"""

        if self._privkey is None:
            return

        privkey_bytes = self.bytes(passphrase, mangleFunction)

        with open(_os.open(filename,
                  _os.O_CREAT | _os.O_WRONLY, 0o700), 'wb') as FILE:
            FILE.write(privkey_bytes)

    def public_key(self, filename=None):
        """Get the public key for this private key. If filename is
           specified then this is written to the passed file"""

        if self._privkey is None:
            return None

        return PublicKey(self._privkey.public_key())

    def key_size_in_bytes(self):
        """Return the number of bytes in this key"""
        if self._privkey is None:
            return 0
        else:
            return int(self._privkey.key_size / 8)

    def fingerprint(self):
        """Return the fingerprint of this key - this is useful to help
           work out which key to use to decrypt data
        """
        return self.public_key().fingerprint()

    def encrypt(self, message):
        """Encrypt and return the passed message"""
        return self.public_key().encrypt(message)

    def verify(self, signature, message):
        """Verify the passed signature is correct for the passed message"""
        return self.public_key().verify(signature, message)

    def decrypt(self, message):
        """Decrypt and return the passed message"""
        key_size = self.key_size_in_bytes()

        if key_size == 0:
            from Acquire.Crypto import DecryptionError
            raise DecryptionError("You cannot decrypt a message "
                                  "with a null key!")

        # try standard decryption
        try:
            message = self._privkey.decrypt(
                message,
                _padding.OAEP(
                    mgf=_padding.MGF1(algorithm=_hashes.SHA256()),
                    algorithm=_hashes.SHA256(),
                    label=None))

            try:
                return message.decode("utf-8")
            except:
                return message
        except:
            pass

        # it is a larger message, so need to decrypt the secret symmetric
        # key, and then use that to decrypt the rest of the token
        try:
            symkey = self._privkey.decrypt(
                        message[0:key_size],
                        _padding.OAEP(
                            mgf=_padding.MGF1(algorithm=_hashes.SHA256()),
                            algorithm=_hashes.SHA256(),
                            label=None))
        except Exception as e:
            from Acquire.Crypto import DecryptionError
            raise DecryptionError(
                "Cannot decrypt the symmetric key used "
                "to encrypt the long message '%s' (%s): %s" %
                (message[0:key_size], key_size, str(e)))

        try:
            f = _fernet.Fernet(symkey)
        except:
            f = _fernet.Fernet(symkey.decode("utf-8"))

        try:
            try:
                message = f.decrypt(message[key_size:])
            except:
                message = f.decrypt(message[key_size:].encode("utf-8"))
        except Exception as e:
            from Acquire.Crypto import DecryptionError
            raise DecryptionError(
                    "Cannot decrypt the long message using the "
                    "symmetric key: %s" % str(e))

        try:
            return message.decode("utf-8")
        except:
            return message

    def sign(self, message):
        """Return the signature for the passed message"""
        if self._privkey is None:
            return None

        if isinstance(message, str):
            message = message.encode("utf-8")

        signature = self._privkey.sign(
                     message,
                     _padding.PSS(
                       mgf=_padding.MGF1(_hashes.SHA256()),
                       salt_length=_padding.PSS.MAX_LENGTH),
                     _hashes.SHA256())

        return signature

    def to_data(self, passphrase, mangleFunction=None):
        """Return the json-serialisable data for this key"""
        data = {}

        b = self.bytes(passphrase, mangleFunction)

        if b is not None:
            data["bytes"] = _bytes_to_string(b)

        return data

    @staticmethod
    def from_data(data, passphrase, mangleFunction=None):
        """Return a private key constructed from the passed json-deserialised
           dictionary
        """

        if isinstance(data, str):
            return PrivateKey.read_bytes(_string_to_bytes(data),
                                         passphrase, mangleFunction)

        elif isinstance(data, bytes):
            return PrivateKey.read_bytes(data, passphrase, mangleFunction)

        elif (data and len(data) > 0):
            return PrivateKey.read_bytes(_string_to_bytes(data["bytes"]),
                                         passphrase, mangleFunction)
        else:
            return None




import uuid as _uuid
from copy import copy as _copy
import datetime as _datetime
import time as _time
import re as _re

__all__ = ["Account"]


def _account_root():
    return "accounting/accounts"


def _get_key_from_date(start, datetime):
    """Return a key encoding the passed date, starting the key with 'start'"""
    from Acquire.ObjectStore import datetime_to_datetime \
        as _datetime_to_datetime
    datetime = _datetime_to_datetime(datetime)
    return "%s/%s" % (start, datetime.date().isoformat())


def _get_date_from_key(key):
    """Return the date that is encoded in the passed key"""
    m = _re.search(r"(\d\d\d\d)-(\d\d)-(\d\d)", key)

    if m:
        from Acquire.ObjectStore import date_and_time_to_datetime \
            as _date_and_time_to_datetime

        return _date_and_time_to_datetime(
                    _datetime.date(year=int(m.groups()[0]),
                                   month=int(m.groups()[1]),
                                   day=int(m.groups()[2])))
    else:
        from Acquire.Accounting import AccountError
        raise AccountError("Could not find a date in the key '%s'" % key)


def _get_datetime_from_key(key):
    """Return the datetime that is encoded in the passed key"""
    m = _re.search(r"(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)\.(\d+)",
                   key)

    if m:
        from Acquire.ObjectStore import date_and_time_to_datetime \
            as _datetime_to_datetime

        return _datetime_to_datetime(
                    _datetime.datetime(year=int(m.groups()[0]),
                                       month=int(m.groups()[1]),
                                       day=int(m.groups()[2]),
                                       hour=int(m.groups()[3]),
                                       minute=int(m.groups()[4]),
                                       second=int(m.groups()[5]),
                                       microsecond=int(m.groups()[6])))
    else:
        from Acquire.Accounting import AccountError
        raise AccountError("Could not find a datetime in the key '%s'" % key)


def _sum_transactions(keys):
    """Internal function that sums all of the transactions identified
        by the passed keys. This returns a tuple of
        (balance, liability, receivable, spent_today)
    """
    from Acquire.Accounting import create_decimal as _create_decimal
    from Acquire.Accounting import TransactionInfo as _TransactionInfo

    balance = _create_decimal(0)
    liability = _create_decimal(0)
    receivable = _create_decimal(0)
    spent_today = _create_decimal(0)

    for key in keys:
        if isinstance(key, _TransactionInfo):
            v = key
        else:
            v = _TransactionInfo(key)

        if v.is_credit():
            balance += v.value()
        elif v.is_debit():
            balance -= v.value()
            spent_today += v.value()
        elif v.is_liability():
            liability += v.value()
            spent_today += v.value()
        elif v.is_accounts_receivable():
            receivable += v.value()
        elif v.is_received_receipt():
            balance -= v.receipted_value()
            liability -= v.value()
        elif v.is_sent_receipt():
            balance += v.receipted_value()
            receivable -= v.value()
        elif v.is_received_refund():
            balance += v.value()
        elif v.is_sent_refund():
            balance -= v.value()

    return (balance, liability, receivable, spent_today)


class Account:
    """This class represents a single account in the ledger. It has a balance,
       and a record of the set of transactions that have been applied.

       The account really holds two accounts: the liability account and
       actual capital account. We combine both together into a single
       account to ensure that updates occur atomically

       All data for this account is stored in the object store
    """
    def __init__(self, name=None, description=None, uid=None, bucket=None):
        """Construct the account. If 'uid' is specified, then load the account
           from the object store (so 'name' and 'description' should be "None")
        """
        if uid is not None:
            self._uid = str(uid)
            self._name = None
            self._description = None
            self._last_update_datetime = None
            self._load_account(bucket)

            if name:
                if name != self.name():
                    from Acquire.Accounting import AccountError
                    raise AccountError(
                        "This account name '%s' does not match what you "
                        "expect! '%s'" % (self.name(), name))

            if description:
                if description != self.description():
                    from Acquire.Accounting import AccountError
                    raise AccountError(
                        "This account description '%s' does not match what "
                        "you expect! '%s'" % (self.description(), description))

        elif name is not None:
            self._uid = None
            self._create_account(name, description)
        else:
            self._uid = None
            self._last_update_datetime = None

    def __str__(self):
        if self._uid is None:
            return "Account::null"
        else:
            return "Account(%s|%s|%s)" % (self._name, self._description,
                                          self._uid)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _create_account(self, name, description):
        """Create the account from scratch"""
        if name is None or description is None:
            from Acquire.Accounting import AccountError
            raise AccountError(
                "You must pass both a name and description to create a new "
                "account")

        if self._uid is not None:
            from Acquire.Accounting import AccountError
            raise AccountError("You cannot create an account twice!")

        from Acquire.Accounting import create_decimal as _create_decimal
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        self._uid = str(_uuid.uuid4())
        self._name = str(name)
        self._description = str(description)
        self._overdraft_limit = _create_decimal(0)
        self._maximum_daily_limit = 0
        self._last_update_datetime = None

        # initialise the account with a balance of zero
        bucket = _get_service_account_bucket()
        self._record_daily_balance(0, 0, 0, bucket=bucket)

        # make sure that this is saved to the object store
        self._save_account(bucket)

    def _get_balance_key(self, datetime=None):
        """Return the balance key for the passed time. This is the key
           into the object store of the object that holds the starting
           balance for the account on the day of the passed datetime.
           If datetime is None, then today's key is returned
        """
        if self.is_null():
            return None

        if datetime is None:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            datetime = _get_datetime_now()
        else:
            from Acquire.ObjectStore import datetime_to_datetime \
                as _datetime_to_datetime
            datetime = _datetime_to_datetime(datetime)

        return _get_key_from_date("%s/balance" % self._key(),
                                  datetime)

    def _record_daily_balance(self, balance, liability, receivable,
                              datetime=None, bucket=None):
        """Record the starting balance for the day containing 'datetime'
           as 'balance' with the starting outstanding liabilities at
           'liability' and starting outstanding accounts receivable at
           'receivable' If 'datetime' is none, then the balance
           for today is set.
        """
        if self.is_null():
            return

        if datetime is None:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            datetime = _get_datetime_now()
        else:
            from Acquire.ObjectStore import datetime_to_datetime \
                as _datetime_to_datetime
            datetime = _datetime_to_datetime(datetime)

        from Acquire.Accounting import create_decimal as _create_decimal

        balance = _create_decimal(balance)
        liability = _create_decimal(liability)
        receivable = _create_decimal(receivable)

        balance_key = self._get_balance_key(datetime)

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        data = {"balance": str(balance),
                "liability": str(liability),
                "receivable": str(receivable)}

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        _ObjectStore.set_object_from_json(bucket, balance_key, data)

    def _reconcile_daily_accounts(self, bucket=None):
        """Internal function used to reconcile the daily accounts.
           This ensures that every line item transaction is summed up
           so that the starting balance for each day is recorded into
           the object store
        """
        if self.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # work back from today to the first day of the account to calculate
        # all of the daily balances... We need to record every day of the
        # account to support quick lookups
        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        today = _get_datetime_now().toordinal()
        day = today
        last_data = None
        num_missing_days = 0

        while last_data is None:
            daytime = _datetime.datetime.fromordinal(day)
            key = self._get_balance_key(daytime)
            last_data = _ObjectStore.get_object_from_json(bucket, key)

            if last_data is None:
                day -= 1
                num_missing_days += 1

                if num_missing_days > 100:
                    # we need another strategy to find the last balance
                    break

        if last_data is None:
            # find the latest day by reading the keys in the object
            # store directly
            root = "%s/balance/" % self._key()
            keys = _ObjectStore.get_all_object_names(
                        bucket, root)

            if keys is None or len(keys) == 0:
                from Acquire.Accounting import AccountError
                raise AccountError(
                    "There is no daily balance recorded for "
                    "the account with UID %s" % self.uid())

            # the encoding of the keys is such that, when sorted, the
            # last key must be the latest balance
            keys.sort()

            last_data = _ObjectStore.get_object_from_json(
                            bucket, keys[-1])
            day = _get_date_from_key(keys[-1]).toordinal()

            if last_data is None:
                raise AccountError("How can there be no data for key %s?" %
                                   keys[-1])

        # what was the balance on the last day?
        from Acquire.Accounting import create_decimal as _create_decimal
        result = (_create_decimal(last_data["balance"]),
                  _create_decimal(last_data["liability"]),
                  _create_decimal(last_data["receivable"]))

        # ok, now we go from the last day until today and sum up the
        # line items from each day to create the daily balances
        # (not including today, as we only want the balance at the beginning
        #  of today)
        for d in range(day+1, today+1):
            day_time = _datetime.datetime.fromordinal(d)
            transaction_keys = self._get_transaction_keys_between(
                            _datetime.datetime.fromordinal(d-1),
                            day_time)

            total = _sum_transactions(transaction_keys)

            result = (result[0]+total[0], result[1]+total[1],
                      result[2]+total[2])

            balance_key = self._get_balance_key(day_time)

            data = {}
            data["balance"] = str(result[0])
            data["liability"] = str(result[1])
            data["receivable"] = str(result[2])

            _ObjectStore.set_object_from_json(bucket, balance_key, data)

    def _get_daily_balance(self, bucket=None, datetime=None):
        """Get the daily starting balance for the passed datetime. This
           returns a tuple of
           (balance, liability, receivable).

           where 'balance' is the current real balance of the account,
           neglecting any outstanding liabilities or accounts receivable,
           where 'liability' is the current total liabilities,
           where 'receivable' is the current total accounts receivable, and

           If datetime is None then todays daily balance is returned. The
           daily balance is the balance at the start of the day. The
           actual balance at a particular time will be this starting
           balance plus/minus all of the transactions between the start
           of that day and the specified datetime
        """
        if self.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        if datetime is None:
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now
            datetime = _get_datetime_now()
        else:
            from Acquire.ObjectStore import datetime_to_datetime \
                as _datetime_to_datetime
            datetime = _datetime_to_datetime(datetime)

        balance_key = self._get_balance_key(datetime)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        data = _ObjectStore.get_object_from_json(bucket, balance_key)

        if data is None:
            # there is no balance for this day. This means that we haven'y
            # yet calculated that day's balance. Do the accounting necessary
            # to construct that day's starting balance
            self._reconcile_daily_accounts(bucket)

            data = _ObjectStore.get_object_from_json(bucket, balance_key)

            if data is None:
                from Acquire.Accounting import AccountError
                raise AccountError("The daily balance for account at date %s "
                                   "is not available" % str(datetime))

        from Acquire.Accounting import create_decimal as _create_decimal

        return (_create_decimal(data["balance"]),
                _create_decimal(data["liability"]),
                _create_decimal(data["receivable"]))

    def _get_balance(self, bucket=None, datetime=None):
        """Get the balance of the account for the passed datetime. This
           returns a tuple of

           (balance, liability, receivable)

           where 'balance' is the current real balance of the account,
           neglecting any outstanding liabilities or accounts receivable,
           where 'liability' is the current total liabilities,
           where 'receivable' is the current total accounts receivable

           If datetime is None then the balance "now" is returned
        """
        if datetime is None:
            return self._get_current_balance(bucket)

        raise NotImplementedError("NOT IMPLEMENTED!")

    def _get_transaction_keys_between(self, start_datetime, end_datetime,
                                      bucket=None):
        """Return all of the object store keys for transactions in this
           account beteen 'start_datetime' and 'end_datetime' (inclusive, e.g.
           start_datetime <= transaction <= end_datetime). This will return an
           empty list if there were no transactions in this time
        """
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # convert both times to UTC
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        start_datetime = _datetime_to_datetime(start_datetime)
        end_datetime = _datetime_to_datetime(end_datetime)

        # get the day for each time
        start_day = start_datetime.toordinal()
        end_day = end_datetime.toordinal()

        if end_datetime.time() == _datetime.time():
            # this ends on midnight of the first day - do not
            # include this last day as nothing will match
            end_day -= 1

        keys = []

        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import date_to_string as _date_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        for day in range(start_day, end_day+1):
            day_date = _datetime.datetime.fromordinal(day)
            day_string = _date_to_string(day_date)

            prefix = "%s/%s" % (self._key(), day_string)
            len_prefix = len(prefix)

            day_keys = _ObjectStore.get_all_object_names(bucket, prefix)

            for day_key in day_keys:
                # the key is Ttime/rest_of_key
                day_key = day_key[len_prefix:]
                time = day_key.split("/")[0]

                datetime = _string_to_datetime("%s%s" % (day_string, time))

                if datetime >= start_datetime and datetime < end_datetime:
                    keys.append("%s/%s" % (prefix, day_key))

        return keys

    def _recalculate_current_balance(self, bucket, now):
        """Internal function that implements _get_current_balance
           by recalculating the total from today from scratch
        """
        # where were we at the start of today?
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        now = _datetime_to_datetime(now)

        (balance, liability, receivable) = self._get_daily_balance(bucket,
                                                                   now)

        # now sum up all of the transactions from today
        from Acquire.ObjectStore import date_and_time_to_datetime \
            as _date_and_time_to_datetime
        start_today = _date_and_time_to_datetime(now.date())

        transaction_keys = self._get_transaction_keys_between(start_today, now)

        total = _sum_transactions(transaction_keys)

        result = (balance+total[0], liability+total[1], receivable+total[2],
                  total[3])

        self._last_update_datetime = now
        self._last_update = result

        return result

    def last_update_datetime(self):
        """Return the last time the balance was updated, as a datetime
           object
        """
        if self._last_update_datetime is None:
            # this is the earliest datetime possible
            from Acquire.ObjectStore import date_and_time_to_datetime \
                as _date_and_time_to_datetime
            return _date_and_time_to_datetime(_datetime.date.fromordinal(1))
        else:
            return self._last_update_datetime

    def _update_current_balance(self, bucket, now):
        """Internal function that implements _get_current_balance
           by updating the balance etc. from transactions that have
           occurred since the last update
        """
        (balance, liability, receivable, spent_today) = self._last_update

        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime

        now = _datetime_to_datetime(now)

        # now sum up all of the transactions since the last update
        transaction_keys = self._get_transaction_keys_between(
                                            self.last_update_datetime(),
                                            now)

        total = _sum_transactions(transaction_keys)

        result = (balance+total[0], liability+total[1], receivable+total[2],
                  spent_today+total[3])

        self._last_update_datetime = now
        self._last_update = result

        return result

    def _get_current_balance(self, bucket=None):
        """Get the balance of the account now (the current balance). This
           returns a tuple of
           (balance, liability, receivable, spent_today).

           where 'balance' is the current real balance of the account,
           neglecting any outstanding liabilities or accounts receivable,
           where 'liability' is the current total liabilities,
           where 'receivable' is the current total accounts receivable, and
           where 'spent_today' is how much has been spent today (from midnight
           until now)
        """
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        now = _get_datetime_now()

        last_update_datetime = self.last_update_datetime()

        if last_update_datetime.date() != now.date():
            # we are on a new day since the last update, so recalculate
            # the balance from scratch
            return self._recalculate_current_balance(bucket, now)
        else:
            # we have calculated the total before today. Get the transactions
            # since the last update and use these to update the daily spend
            # etc.
            return self._update_current_balance(bucket, now)

    def is_null(self):
        """Return whether or not this is a null account"""
        return self._uid is None

    def name(self):
        """Return the name of this account"""
        if self.is_null():
            return None

        return self._name

    def description(self):
        """Return the description of this account"""
        if self.is_null():
            return None

        return self._description

    def uid(self):
        """Return the UID for this account."""
        return self._uid

    def _key(self):
        """Return the key for this account in the object store"""
        if self.is_null():
            return None

        return "%s/%s" % (_account_root(), self.uid())

    def _load_account(self, bucket=None):
        """Load the current state of the account from the object store"""
        if self.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        data = _ObjectStore.get_object_from_json(bucket, self._key())
        self.__dict__ = _copy(Account.from_data(data).__dict__)

    def _save_account(self, bucket=None):
        """Save this account back to the object store"""
        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        _ObjectStore.set_object_from_json(bucket, self._key(), self.to_data())

    def to_data(self):
        """Return a dictionary that can be encoded to json from this object"""
        data = {}

        if not self.is_null():
            data["uid"] = self._uid
            data["name"] = self._name
            data["description"] = self._description
            data["overdraft_limit"] = str(self._overdraft_limit)
            data["maximum_daily_limit"] = str(self._maximum_daily_limit)

        return data

    @staticmethod
    def from_data(data):
        """Construct and return an Account from the passed dictionary that has
           been decoded from json
        """
        account = Account()

        if (data and len(data) > 0):
            from Acquire.Accounting import create_decimal as _create_decimal

            account._uid = data["uid"]
            account._name = data["name"]
            account._description = data["description"]
            account._overdraft_limit = _create_decimal(data["overdraft_limit"])
            account._maximum_daily_limit = _create_decimal(
                                                data["maximum_daily_limit"])

        return account

    def assert_valid_authorisation(self, authorisation):
        """Assert that the passed authorisation is valid for this
           account
        """
        from Acquire.Identity import Authorisation as _Authorisation
        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The passed authorisation must be an "
                            "Authorisation")

    def _get_safe_now(self):
        """This function returns the current time. It avoids dangerous
           times (when the system may be updating) by sleeping through
           those times
        """
        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now
        now = _get_datetime_now()

        # don't allow any transactions in the last 30 seconds of the day, as we
        # will sum up the day balance at midnight, and don't want to risk any
        # late transactions from messing up the accounting
        while now.hour == 23 and now.minute == 59 and now.second >= 30:
            _time.sleep(5)
            now = _get_datetime_now()

        return now

    def _delete_note(self, note, bucket=None):
        """Internal function called to delete the passed note from the
           record. This is unsafe and should only be called from
           DebitNote.return_value or CreditNote.return_value (which
           themselves are only called from Ledger)
        """
        if note is None:
            return

        from Acquire.Accounting import DebitNote as _DebitNote
        from Acquire.Accounting import CreditNote as _CreditNote

        if isinstance(note, _DebitNote) or isinstance(note, _CreditNote):
            item_key = "%s/%s" % (self._key(), note.uid())

            if bucket is None:
                from Acquire.Service import get_service_account_bucket \
                    as _get_service_account_bucket
                bucket = _get_service_account_bucket()

            # remove the note
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            from Acquire.ObjectStore import get_datetime_now \
                as _get_datetime_now

            try:
                _ObjectStore.delete_object(bucket, item_key)
            except:
                pass

            # now remove all day-balances from the day before this note
            # to today. Hopefully this will prevent any ledger errors...
            day0 = note.datetime().toordinal() - 1
            day1 = _get_datetime_now().toordinal()

            for day in range(day0, day1+1):
                balance_key = self._get_balance_key(
                                    _datetime.datetime.fromordinal(day))

                try:
                    _ObjectStore.delete_object(bucket, balance_key)
                except:
                    pass

    def _credit_refund(self, debit_note, refund, bucket=None):
        """Credit the value of the passed 'refund' to this account. The
           refund must be for a previous completed debit, hence the
           original debitted value is returned to the account.
        """
        from Acquire.Accounting import Refund as _Refund
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(refund, _Refund):
            raise TypeError("The passed refund must be a Refund")

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if refund.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        if refund.value() != debit_note.value():
            raise ValueError("The refunded value does not match the value "
                             "of the debit note: %s versus %s" %
                             (refund.value(), debit_note.value()))

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                        _TransactionCode.RECEIVED_REFUND,
                                        refund.value())

        # create a UID and datetime for this credit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this credit later. The key is made
        # up from the iso format of the datetime of the credit
        # and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)
        l = _LineItem(debit_note.uid(), refund.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit_refund(self, refund, bucket=None):
        """Debit the value of the passed 'refund' from this account. The
           refund must be for a previous completed credit. There is a risk
           that this value has been spent, so this is one of the only
           functions that allows a balance to drop below an overdraft or
           other limit (as the refund should always succeed).
        """
        from Acquire.Accounting import Refund as _Refund

        if not isinstance(refund, _Refund):
            raise TypeError("The passed refund must be a Refund")

        if refund.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(_TransactionCode.SENT_REFUND,
                                                refund.value())

        # create a UID and datetime for this debit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this debit later. The key is made
        # up from the date and  of the debit and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)
        l = _LineItem(uid, refund.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _credit_receipt(self, debit_note, receipt, bucket=None):
        """Credit the value of the passed 'receipt' to this account. The
           receipt must be for a previous provisional credit, hence the
           money is awaiting transfer from accounts receivable.
        """
        from Acquire.Accounting import Receipt as _Receipt
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(receipt, _Receipt):
            raise TypeError("The passed receipt must be a Receipt")

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if receipt.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        if receipt.receipted_value() != debit_note.value():
            raise ValueError("The receipted value does not match the value "
                             "of the debit note: %s versus %s" %
                             (receipt.receipted_value(), debit_note.value()))

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.SENT_RECEIPT,
                                    receipt.value(), receipt.receipted_value())

        # create a UID and datetime for this credit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this credit later. The key is made
        # up from the isoformat datetime of the credit and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)
        l = _LineItem(debit_note.uid(), receipt.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit_receipt(self, receipt, bucket=None):
        """Debit the value of the passed 'receipt' from this account. The
           receipt must be for a previous provisional debit, hence
           the money should be available.
        """
        from Acquire.Accounting import Receipt as _Receipt

        if not isinstance(receipt, _Receipt):
            raise TypeError("The passed receipt must be a Receipt")

        if receipt.is_null():
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        encoded_value = _TransactionInfo.encode(
                                    _TransactionCode.RECEIVED_RECEIPT,
                                    receipt.value(), receipt.receipted_value())

        # create a UID and datetime for this debit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this debit later. The key is made
        # up from the isoformat datetime of the debit and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)
        l = _LineItem(uid, receipt.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _credit(self, debit_note, bucket=None):
        """Credit the value in 'debit_note' to this account. If the debit_note
           shows that the payment is provisional then this will be recorded
           as accounts receivable. This will record the credit with the
           same UID as the debit identified in the debit_note, so that
           we can reconcile all credits against matching debits.
        """
        from Acquire.Accounting import DebitNote as _DebitNote

        if not isinstance(debit_note, _DebitNote):
            raise TypeError("The passed debit note must be a DebitNote")

        if debit_note.value() <= 0:
            return

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        if debit_note.is_provisional():
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.ACCOUNT_RECEIVABLE,
                                debit_note.value())
        else:
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.CREDIT,
                                debit_note.value())

        # create a UID and datetime for this credit and record
        # it in the account
        now = self._get_safe_now()

        # and to create a key to find this credit later. The key is made
        # up from the isoformat datetime of the credit and a random string
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)

        # the line item records the UID of the debit note, so we can
        # find this debit note in the system and, from this, get the
        # original transaction in the transaction record
        l = _LineItem(debit_note.uid(), debit_note.authorisation())

        _ObjectStore.set_object_from_json(bucket, item_key, l.to_data())

        return (uid, now)

    def _debit(self, transaction, authorisation,
               is_provisional, receipt_by, bucket=None):
        """Debit the value of the passed transaction from this account based
           on the authorisation contained
           in 'authorisation'. This will create a unique ID (UID) for
           this debit and will return this together with the datetime of the
           debit. If this transaction 'is_provisional' then it will be
           recorded as a liability, which must be receipted before
           'receipt_by'. If 'receipt_by' is None, then this will
           automatically be 1 week in the future

           The UID will encode both the date of the debit and provide a random
           ID that together can be used to identify the transaction associated
           with this debit in the future.

           This will raise an exception if the debit cannot be completed, e.g.
           if the authorisation is invalid, if the debit exceeds a limit or
           there are insufficient funds in the account

           Note that this function is private as it should only be called
           by the DebitNote class
        """
        if self.is_null() or transaction.value() <= 0:
            return None

        from Acquire.Accounting import Transaction as _Transaction

        if not isinstance(transaction, _Transaction):
            raise TypeError("The passed transaction must be a Transaction!")

        self.assert_valid_authorisation(authorisation)

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        if self.available_balance(bucket) < transaction.value():
            from Acquire.Accounting import InsufficientFundsError
            raise InsufficientFundsError(
                "You cannot debit '%s' from account %s as there "
                "are insufficient funds in this account." %
                (transaction, str(self)))

        # create a UID and datetime for this debit and record
        # it in the account
        now = self._get_safe_now()

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import datetime_to_datetime \
            as _datetime_to_datetime
        from Acquire.ObjectStore import get_datetime_future \
            as _get_datetime_future
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import LineItem as _LineItem

        from Acquire.Accounting import TransactionInfo as _TransactionInfo
        from Acquire.Accounting import TransactionCode as _TransactionCode

        if is_provisional:
            if receipt_by is None:
                receipt_by = _get_datetime_future(days=7)
            else:
                receipt_by = _datetime_to_datetime(receipt_by)

            delta = (receipt_by - now).total_seconds()
            if delta < 3600:
                from Acquire.Accounting import AccountError
                raise AccountError(
                    "You cannot request a receipt to be provided less "
                    "than 1 hour into the future! %s versus %s is only "
                    "%s second(s) in the future!" %
                    (_datetime_to_string(receipt_by),
                     _datetime_to_string(now), delta))
        else:
            receipt_by = None

        # and to create a key to find this debit later. The key is made
        # up from the isoformat datetime of the debit and a random string
        day_key = _datetime_to_string(now)
        uid = "%s/%s" % (day_key, str(_uuid.uuid4())[0:8])

        # the key in the object store is a combination of the key for this
        # account plus the uid for the debit plus the actual debit value.
        # We record the debit value in the key so that we can accumulate
        # the balance from just the key names
        if is_provisional:
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.CURRENT_LIABILITY,
                                transaction.value())
        else:
            encoded_value = _TransactionInfo.encode(
                                _TransactionCode.DEBIT,
                                transaction.value())

        item_key = "%s/%s/%s" % (self._key(), uid, encoded_value)

        # create a line_item for this debit and save it to the object store
        line_item = _LineItem(uid, authorisation)

        _ObjectStore.set_object_from_json(bucket, item_key,
                                          line_item.to_data())

        if self.is_beyond_overdraft_limit(bucket):
            # This transaction has helped push the account beyond the
            # overdraft limit. Delete the transaction and raise
            # an InsufficientFundsError

            _ObjectStore.delete_object(bucket, item_key)
            raise InsufficientFundsError(
                "You cannot debit '%s' from account %s as there "
                "are insufficient funds in this account." %
                (transaction, str(self)))

        return (uid, now, receipt_by)

    def available_balance(self, bucket=None):
        """Return the available balance of this account. This is the amount
           of value that can be spent (e.g. includes overdraft and fixed daily
           spend limits, and except any outstanding liabilities)
        """
        result = self._get_current_balance(bucket)

        balance = result[0]
        liabilities = result[1]
        spent_today = result[2]

        available = balance - liabilities + self.get_overdraft_limit()

        if self._maximum_daily_limit:
            return min(available, self._maximum_daily_limit - spent_today)
        else:
            return available

    def balance(self, bucket=None):
        """Return the current balance of this account"""
        result = self._get_current_balance(bucket)
        return result[0]

    def liability(self, bucket=None):
        """Return the current total liability of this account"""
        result = self._get_current_balance(bucket)
        return result[1]

    def receivable(self, bucket=None):
        """Return the current total accounts receivable of this account"""
        result = self._get_current_balance(bucket)
        return result[2]

    def spent_today(self, bucket=None):
        """Return the current amount spent today on this account"""
        result = self._get_current_balance(bucket)
        return result[3]

    def balance_status(self, bucket=None):
        """Return the overall balance status as a dictionary
           with keys 'balance', 'liability', 'receivable' and
           'spent_today'
        """
        result = self._get_current_balance(bucket)
        d = {}
        d["balance"] = result[0]
        d["liability"] = result[1]
        d["receivable"] = result[2]
        d["spent_today"] = result[3]
        return d

    def get_overdraft_limit(self):
        """Return the overdraft limit of this account"""
        if self.is_null():
            return 0

        return self._overdraft_limit

    def set_overdraft_limit(self, limit, bucket=None):
        """Set the overdraft limit of this account to 'limit'"""
        if self.is_null():
            return

        from Acquire.Accounting import create_decimal as _create_decimal
        limit = _create_decimal(limit)
        if limit < 0:
            raise ValueError("You cannot set the overdraft limit to a "
                             "negative value! (%s)" % limit)

        old_limit = self._overdraft_limit

        if old_limit != limit:
            self._overdraft_limit = limit

            if self.is_beyond_overdraft_limit():
                # restore the old limit
                self._overdraft_limit = old_limit
                from Acquire.Accounting import AccountError
                raise AccountError("You cannot change the overdraft limit to "
                                   "%s as this is greater than the current "
                                   "balance!" % (limit))
            else:
                # save the new limit to the object store
                self._save_account(bucket)

    def is_beyond_overdraft_limit(self, bucket=None):
        """Return whether or not the current balance is beyond
           the overdraft limit
        """
        result = self._get_current_balance(bucket)

        return (result[0] - result[1]) < -(self.get_overdraft_limit())




import io as _io
import datetime as _datetime
import uuid as _uuid
import json as _json
import os as _os

__all__ = ["ObjectStore", "set_object_store_backend",
           "use_testing_object_store_backend",
           "use_oci_object_store_backend"]

_objstore_backend = None


def use_testing_object_store_backend(backend):
    from ._testing_objstore import Testing_ObjectStore as _Testing_ObjectStore
    set_object_store_backend(_Testing_ObjectStore)
    bucket = "%s/testing_objstore" % backend

    try:
        # make sure that this directory exists if it doesn't already
        _os.mkdir(bucket)
    except:
        pass

    return bucket


def use_oci_object_store_backend():
    from ._oci_objstore import OCI_ObjectStore as _OCI_ObjectStore
    set_object_store_backend(_OCI_ObjectStore)


class ObjectStore:
    @staticmethod
    def create_bucket(bucket, bucket_name, compartment=None):
        """Create and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. This will raise an
           ObjectStoreError if this bucket already exists
        """
        return _objstore_backend.create_bucket(bucket, bucket_name,
                                               compartment)

    @staticmethod
    def get_bucket(bucket, bucket_name, compartment=None,
                   create_if_needed=True):
        """Find and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. If 'create_if_needed' is True
           then the bucket will be created if it doesn't exist. Otherwise,
           if the bucket does not exist then an exception will be raised.
        """
        return _objstore_backend.get_bucket(bucket, bucket_name,
                                            compartment, create_if_needed)

    @staticmethod
    def create_par(bucket, encrypt_key, key=None, readable=True,
                   writeable=False, duration=3600):
        """Create a pre-authenticated request for the passed bucket and
           key (if key is None then the request is for the entire bucket).
           This will return a PAR object that will contain a URL that can
           be used to access the object/bucket. If writeable is true, then
           the URL will also allow the object/bucket to be written to.
           PARs are time-limited. Set the lifetime in seconds by passing
           in 'duration' (by default this is one hour). Note that you must
           pass in a public key that will be used to encrypt this PAR. This is
           necessary as the PAR grants access to anyone who can decrypt
           the URL
        """
        from Acquire.Client import PAR as _PAR

        par = _objstore_backend.create_par(bucket, encrypt_key, key, readable,
                                           writeable, duration)

        if not isinstance(par, _PAR):
            raise TypeError("A create_par request should always return an "
                            "value of type PAR: %s is not correct!" % par)

        return par

    @staticmethod
    def delete_par(bucket, par):
        """Delete the passed PAR, which provides access to data in the
           passed bucket
        """
        _objstore_backend.delete_par(bucket=bucket, par=par)

    @staticmethod
    def get_object_as_file(bucket, key, filename):
        """Get the object contained in the key 'key' in the passed 'bucket'
           and writing this to the file called 'filename'"""
        return _objstore_backend.get_object_as_file(bucket, key, filename)

    @staticmethod
    def get_object(bucket, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket"""
        return _objstore_backend.get_object(bucket, key)

    @staticmethod
    def get_string_object(bucket, key):
        """Return the string in 'bucket' associated with 'key'"""
        return _objstore_backend.get_string_object(bucket, key)

    @staticmethod
    def get_object_from_json(bucket, key):
        """Return an object constructed from json stored at 'key' in
           the passed bucket. This returns None if there is no data
           at this key
        """
        return _objstore_backend.get_object_from_json(bucket, key)

    @staticmethod
    def get_all_object_names(bucket, prefix=None):
        """Returns the names of all objects in the passed bucket"""
        return _objstore_backend.get_all_object_names(bucket, prefix)

    @staticmethod
    def get_all_objects(bucket, prefix=None):
        """Return all of the objects in the passed bucket"""
        return _objstore_backend.get_all_objects(bucket, prefix)

    @staticmethod
    def get_all_objects_from_json(bucket, prefix=None):
        """Return all of the objects in the passed bucket as
           json-deserialised objects
        """
        return _objstore_backend.get_all_objects_from_json(bucket, prefix)

    @staticmethod
    def get_all_strings(bucket, prefix=None):
        """Return all of the strings in the passed bucket"""
        return _objstore_backend.get_all_strings(bucket, prefix)

    @staticmethod
    def set_object(bucket, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""
        _objstore_backend.set_object(bucket, key, data)

    @staticmethod
    def set_object_from_file(bucket, key, filename):
        """Set the value of 'key' in 'bucket' to equal the contents
           of the file located by 'filename'"""
        _objstore_backend.set_object_from_file(bucket, key, filename)

    @staticmethod
    def set_ins_object_from_json(bucket, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json, if (and only if)
           this key has not already been set (ins = 'if not set').
           This returns the object at this key after the operation
           (either the set object or the value that was previously
           set
        """
        from Acquire.ObjectStore import Mutex as _Mutex
        m = _Mutex(bucket=bucket, key=key)

        try:
            old_data = ObjectStore.get_object_from_json(bucket, key)
        except:
            old_data = None

        if old_data is not None:
            m.unlock()
            return old_data
        else:
            try:
                ObjectStore.set_object_from_json(bucket, key, data)
            except:
                m.unlock()
                raise

            return data

    @staticmethod
    def set_ins_string_object(bucket, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data',
           if (and only if) this key has not already been set
           (ins = 'if not set'). This returns the object at this
           key after the operation (either the set string, or the value
           that was previously set)
        """
        from Acquire.ObjectStore import Mutex as _Mutex
        m = _Mutex(bucket=bucket, key=key)

        try:
            val = ObjectStore.get_string_object(bucket, key)
        except:
            val = None

        if val is not None:
            m.unlock()
            return val
        else:
            try:
                ObjectStore.set_string_object(bucket, key, string_data)
            except:
                m.unlock()
                raise

            return string_data

    @staticmethod
    def set_string_object(bucket, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data'"""
        _objstore_backend.set_string_object(bucket, key, string_data)

    @staticmethod
    def set_object_from_json(bucket, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json"""
        _objstore_backend.set_object_from_json(bucket, key, data)

    @staticmethod
    def delete_all_objects(bucket, prefix=None):
        """Deletes all objects..."""
        _objstore_backend.delete_all_objects(bucket, prefix)

    @staticmethod
    def delete_object(bucket, key):
        """Removes the object at 'key'"""
        _objstore_backend.delete_object(bucket, key)

    @staticmethod
    def clear_all_except(bucket, keys):
        """Removes all objects from the passed 'bucket' except those
           whose keys are or start with any key in 'keys'"""
        _objstore_backend.clear_all_except(bucket, keys)

    @staticmethod
    def get_size_and_checksum(bucket, key):
        """Return the object size (in bytes) and checksum of the
           object in the passed bucket at the specified key
        """
        return _objstore_backend.get_size_and_checksum(bucket, key)


def set_object_store_backend(backend):
    """Set the backend that is used to actually connect to
       the object store. This can only be set once in the program!
    """
    global _objstore_backend

    if backend == _objstore_backend:
        return

    if _objstore_backend is not None:
        from Acquire.ObjectStore import ObjectStoreError
        raise ObjectStoreError("You cannot change the object store "
                               "backend once it has been already set!")

    _objstore_backend = backend




__all__ = ["clear_service_cache"]


def clear_service_cache():
    """Call this function to clear the cache of the current service.
       This is only really needed for testhing, when a single python
      interpreter will move between multiple (cached) services
    """
    from Acquire.Service import clear_services_cache
    from Acquire.Service import clear_serviceinfo_cache, clear_login_cache

    clear_services_cache()
    clear_login_cache()
    clear_serviceinfo_cache()




import json as _json
from io import BytesIO as _BytesIO

__all__ = ["call_function", "pack_arguments", "unpack_arguments",
           "create_return_value", "pack_return_value", "unpack_return_value",
           "exception_to_safe_exception", "exception_to_string"]


def _get_signing_certificate(fingerprint=None):
    """Return the signing certificate for this service"""
    from ._service_account import get_service_private_certificate \
        as _get_service_private_certificate
    return _get_service_private_certificate(fingerprint=fingerprint)


def _get_key(key, fingerprint=None):
    """The user may pass the key in multiple ways. It could just be
       a key. Or it could be a function that gets the key on demand.
       Or it could be a dictionary that has the key stored under
       "encryption_public_key"
    """
    from Acquire.Crypto import PublicKey as _PublicKey
    from Acquire.Crypto import PrivateKey as _PrivateKey

    if key is None:
        return None
    elif isinstance(key, _PublicKey) or isinstance(key, _PrivateKey):
        key = key
    elif isinstance(key, dict):
        try:
            key = key["encryption_public_key"]
        except:
            key = None

        if key is not None:
            from Acquire.ObjectStore import string_to_bytes as _string_to_bytes
            key = _PublicKey.read_bytes(_string_to_bytes(key))
    else:
        key = key(fingerprint=fingerprint)

    if fingerprint is not None:
        if key is None:
            from Acquire.Crypto import KeyManipulationError
            raise KeyManipulationError(
                "Cannot find the key with fingerprint %s!" % fingerprint)
        elif key.fingerprint() != fingerprint:
            from Acquire.Crypto import KeyManipulationError
            raise KeyManipulationError(
                "Cannot find a key with the required fingerprint (%s). "
                "The only key has fingerprint %s" %
                (fingerprint, key.fingerprint()))

    return key


def create_return_value(status=0, message="success", log=None, error=None):
    """Convenience functiont that creates the start of the
       return_value dictionary, setting the "status" key
       to the value of 'status', the "message" key to the
       value of 'message' and the "log" key to the value
       of 'log'. If 'error' is passed, then this signifies
       an exception, which will be packed and returned.
       This returns a simple dictionary, which
       is ready to be packed into a json string
    """

    if error:
        if isinstance(error, Exception):
            import traceback as _traceback

            status = -2
            message = "%s: %s\nTraceback:\n%s" % (
                str(error.__class__.__name__),
                " : ".join(error.args),
                "".join(_traceback.format_tb(error.__traceback__)))
        else:
            status = -1
            message = str(error)

    return_value = {}
    return_value["status"] = status
    return_value["message"] = message

    if log:
        return_value["log"] = log

    return return_value


def pack_return_value(result, key=None, response_key=None, public_cert=None):
    """Pack the passed result into a json string, optionally
       encrypting the result with the passed key, and optionally
       supplying a public response key, with which the function
       being called should encrypt the response. If public_cert is
       provided then we will ask the service to sign their response.
       Note that you can only ask the service to sign their response
       if you provide a 'reponse_key' for them to encrypt it with too
    """

    try:
        sign_result = key["sign_with_service_key"]
    except:
        sign_result = False

    key = _get_key(key)
    response_key = _get_key(response_key)

    from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
    from Acquire.ObjectStore import get_datetime_now_to_string \
        as _get_datetime_now_to_string

    if response_key:
        result["encryption_public_key"] = _bytes_to_string(
                                            response_key.bytes())

        if public_cert:
            result["sign_with_service_key"] = public_cert.fingerprint()

    elif sign_result and (key is None):
        from Acquire.Service import PackingError
        raise PackingError(
            "You cannot ask the service to sign the response "
            "without also providing a key to encrypt it with too")

    if key is None:
        result["synctime"] = _get_datetime_now_to_string()

    result = _json.dumps(result).encode("utf-8")

    if key is not None:
        response = {}

        result_data = key.encrypt(result)

        if sign_result:
            # sign using the signing certificate for this service
            signature = _get_signing_certificate(
                            fingerprint=sign_result).sign(result_data)
            response["signature"] = _bytes_to_string(signature)

        response["data"] = _bytes_to_string(result_data)
        response["encrypted"] = True
        response["fingerprint"] = key.fingerprint()
        response["synctime"] = _get_datetime_now_to_string()
        result = _json.dumps(response).encode("utf-8")

    elif sign_result:
        from Acquire.Service import PackingError
        raise PackingError(
            "The service must encrypt the response before it "
            "can be signed.")

    return result


def pack_arguments(args, key=None, response_key=None, public_cert=None):
    """Pack the passed arguments, optionally encrypted using the passed key"""
    return pack_return_value(args, key, response_key, public_cert)


def exception_to_safe_exception(e):
    """Convert the passed exception to a "safe" exception - this is one
       that can be copied because it does not hold references to any
       local data
    """
    if not issubclass(e.__class__, Exception):
        return TypeError(str(e))

    import tblib as _tblib
    tb = _tblib.Traceback(e.__traceback__)
    e.__traceback__ = tb.as_traceback()

    return e


def unpack_arguments(args, key=None, public_cert=None, is_return_value=False,
                     function=None, service=None):
    """Call this to unpack the passed arguments that have been encoded
       as a json string, packed using pack_arguments. This will always
       return a dictionary. If there are no arguments, then an empty
       dictionary will be returned. If 'public_cert' is supplied then
       a signature of the result will be verified using 'public_cert'

       This function is also called as unpack_return_value, in which
       case is_return_value is set as True. The 'function' on 'service'
       that was called (or to be called) can also be passed. These
       are used to help provide more context for error messages.
    """
    if not (args and len(args) > 0):
        return {}

    # args should be a json-encoded utf-8 string
    try:
        data = _json.loads(args)
    except Exception as e:
        from Acquire.Service import UnpackingError
        raise UnpackingError("Cannot decode json from '%s' : %s" %
                             (data, str(e)))

    while not isinstance(data, dict):
        if not (data and len(data) > 0):
            return {}

        try:
            data = _json.loads(data)
        except Exception as e:
            from Acquire.Service import UnpackingError
            raise UnpackingError(
                "Cannot decode a json dictionary from '%s' : %s" %
                (data, str(e)))

    if is_return_value:
        # extra checks if this is a return value of a function rather
        # than the arguments
        if len(data) == 1 and "error" in data:
            from Acquire.Service import RemoteFunctionCallError
            raise RemoteFunctionCallError(
                "Calling %s on %s resulted in error: '%s'" %
                (function, service, data["error"]))

        if "status" in data:
            if data["status"] != 0:
                if "exception" in data:
                    _unpack_and_raise(function, service, data["exception"])
                else:
                    from Acquire.Service import RemoteFunctionCallError
                    raise RemoteFunctionCallError(
                        "Calling %s on %s exited with status %d: %s" %
                        (function, service, data["status"], data["message"]))

    try:
        is_encrypted = data["encrypted"]
    except:
        is_encrypted = False

    from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

    if public_cert:
        if not is_encrypted:
            raise UnpackingError(
                "Cannot unpack the result of %s on %s as it should be "
                "signed, but it isn't! (only encrypted results are signed) "
                "Response == %s" % (function, service, _json.dumps(data)))

        try:
            signature = _string_to_bytes(data["signature"])
        except:
            signature = None

        if signature is None:
            raise UnpackingError(
                "We requested that the data was signed "
                "when calling %s on %s, but a signature was not provided!" %
                (function, service))

    if is_encrypted:
        encrypted_data = _string_to_bytes(data["data"])

        try:
            fingerprint = data["fingerprint"]
        except:
            fingerprint = None

        if public_cert:
            try:
                public_cert.verify(signature, encrypted_data)
            except Exception as e:
                raise UnpackingError(
                    "The signature of the returned data "
                    "from calling %s on %s "
                    "is incorrect and does not match what we "
                    "know! %s" % (function, service, str(e)))

        decrypted_data = _get_key(key, fingerprint).decrypt(encrypted_data)
        return unpack_arguments(decrypted_data,
                                is_return_value=is_return_value,
                                function=function, service=service)
    else:
        return data


def unpack_return_value(return_value, key=None, public_cert=None,
                        function=None, service=None):
    """Call this to unpack the passed arguments that have been encoded
       as a json string, packed using pack_arguments"""
    return unpack_arguments(return_value, key, public_cert,
                            is_return_value=True, function=function,
                            service=service)


class _UnicodeDecodeError(Exception):
    """Fake UnicodeDecodeError as the real one has a strange
       constructor syntax
    """
    pass


def _unpack_and_raise(function, service, exdata):
    """This function unpacks the exception whose data is in 'exdata',
       and raises it in the current thread. Additional information
       is added to the error message to include the remote function
       that was called (function) and the service on which it
       was called.

       The exdata should be a dictionary containing:

       class: class name of the exception
       module: module containing the exception
       traceback: json-serialised traceback (dumped using tbblib)
       error: error message of the exception
    """
    try:
        import importlib as _importlib
        import tblib as _tblib

        if exdata["class"] == "UnicodeDecodeError":
            exclass = _UnicodeDecodeError
        else:
            mod = _importlib.import_module(exdata["module"])
            exclass = getattr(mod, exdata["class"])

        ex = exclass("Error calling '%s' on '%s': %s" %
                     (function, service, exdata["error"]))
        ex.__traceback__ = _tblib.Traceback.from_dict(
                                    exdata["traceback"]).as_traceback()
    except Exception as e:
        from Acquire.Service import RemoteFunctionCallError
        from Acquire.Service import exception_to_string \
            as _exception_to_string
        raise RemoteFunctionCallError(
            "An exception occurred while calling '%s' on '%s'\n\n"
            "CAUSE: %s\n\nEXDATA: %s" %
            (function, service, _exception_to_string(e), exdata))

    raise ex


def exception_to_string(e):
    """This function returns a simple string that represents the exception,
       including the first line of the traceback.
    """
    import traceback as _traceback
    lines = _traceback.format_exception(e.__class__, e, e.__traceback__,
                                        limit=2)

    return "".join(lines)


def _call_local_function(service, function=None, args_key=None,
                         response_key=None, public_cert=None, args=None):
    """This is an internal version of call_function which short-cuts
       the whole process if the function is being called in the local
       service
    """
    response_key = _get_key(response_key)

    if function is not None:
        args["function"] = function

    if response_key:
        args_json = pack_arguments(args, args_key, response_key.public_key(),
                                   public_cert=public_cert)
    else:
        args_json = pack_arguments(args, args_key)

    result = service._call_local_function(function, args_json)

    # Now unpack the results
    return unpack_return_value(result, response_key, public_cert,
                               function=function, service=service)


def call_function(service_url, function=None, args_key=None, response_key=None,
                  public_cert=None, args=None, **kwargs):
    """Call the remote function called 'function' at 'service_url' passing
       in named function arguments in 'kwargs'. If 'args_key' is supplied,
       then encrypt the arguments using 'args'. If 'response_key'
       is supplied, then tell the remote server to encrypt the response
       using the public version of 'response_key', so that we can
       decrypt it in the response. If 'public_cert' is supplied then
       we will ask the service to sign their response using their
       service signing certificate, and we will validate the
       signature using 'public_cert'
    """
    if args is None:
        args = {}

    for key, value in kwargs.items():
        args[key] = value

    from Acquire.Service import is_running_service as _is_running_service
    from Acquire.ObjectStore import get_datetime_now_to_string \
        as _get_datetime_now_to_string

    service = None

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        try:
            service = _get_this_service(need_private_access=True)
        except:
            pass

        if service is not None:
            if service.canonical_url() == service_url:
                return _call_local_function(service, function, args_key,
                                            response_key, public_cert, args)

    response_key = _get_key(response_key)

    if function is not None:
        args["function"] = function

    if response_key:
        args_json = pack_arguments(args, args_key, response_key.public_key(),
                                   public_cert=public_cert)
    else:
        args_json = pack_arguments(args, args_key)

    response = None
    try:
        from Acquire.Stubs import requests as _requests
        #print("SENDING '%s'" % args_json)
        response = _requests.post(service_url, data=args_json, timeout=15.0)
        #print("RECEIVED '%s'" % response)
    except Exception as e:
        from Acquire.Service import RemoteFunctionCallError
        raise RemoteFunctionCallError(
            "Cannot call remote function '%s' at '%s' because of a possible "
            "network issue: requests exeption = '%s'" %
            (function, service_url, str(e)))

    args = None
    args_json = None
    args_key = None

    if response.status_code != 200:
        from Acquire.Service import RemoteFunctionCallError
        raise RemoteFunctionCallError(
            "Cannot call remote function '%s' as '%s'. Invalid error code "
            "%d returned. Message:\n%s" %
            (function, service_url,
             response.status_code, str(response.content)))

    if response.encoding == "utf-8" or response.encoding is None:
        result = response.content.decode("utf-8")
        #print("RECEIVED DATA '%s'" % result)
    else:
        from Acquire.Service import RemoteFunctionCallError
        raise RemoteFunctionCallError(
            "Cannot call remote function '%s' as '%s'. Invalid data encoding "
            "%s returned. Message:\n%s" %
            (function, service_url,
             response.encoding, str(response.content)))

    return unpack_return_value(result, response_key, public_cert,
                               function=function, service=service_url)




__all__ = ["trust_service", "untrust_service"]


def trust_service(service, authorisation=None):
    """Trust the passed service. This will record this service as trusted,
       e.g. saving the keys and certificates for this service and allowing
       it to be used for the specified type. You must pass in a valid
       admin_user authorisation for this service
    """
    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        from Acquire.Identity import Authorisation as _Authorisation

        if not isinstance(authorisation, _Authorisation):
            raise PermissionError(
                "You must supply a valid authorisation when you want to trust "
                "a new service!")

        local_service = _get_this_service(need_private_access=True)
        local_service.assert_admin_authorised(
            authorisation, "trust_service %s" % service.uid())

        bucket = _get_service_account_bucket()

        urlkey = "_trusted/url/%s" % _url_to_encoded(service.canonical_url())
        uidkey = "_trusted/uid/%s" % service.uid()
        service_data = service.to_data()

        # store the trusted service by both canonical_url and uid
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        _ObjectStore.set_object_from_json(bucket, uidkey, service_data)
        _ObjectStore.set_string_object(bucket, urlkey, uidkey)

        from Acquire.Service import clear_services_cache \
            as _clear_services_cache
        _clear_services_cache()
    else:
        from Acquire.Client import Wallet as _Wallet
        _Wallet.add_service(service)


def untrust_service(service, authorisation=None):
    """Stop trusting the passed service. This will remove the service
       as being trusted. You must pass in a valid admin_user authorisation
       for this service
    """
    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        from Acquire.Identity import Authorisation as _Authorisation

        if not isinstance(authorisation, _Authorisation):
            raise PermissionError(
                "You must supply a valid authorisation when you want to "
                "stop trusting a service!")

        local_service = _get_this_service(need_private_access=True)
        local_service.assert_admin_authorised(
            authorisation, "untrust_service %s" % service.uid())

        bucket = _get_service_account_bucket()
        urlkey = "_trusted/url/%s" % _url_to_encoded(service.canonical_url())
        uidkey = "_trusted/uid/%s" % service.uid()

        # delete the trusted service by both canonical_url and uid
        try:
            _ObjectStore.delete_object(bucket, uidkey)
        except:
            pass

        try:
            _ObjectStore.delete_object(bucket, urlkey)
        except:
            pass

        from Acquire.Service import clear_services_cache \
            as _clear_services_cache
        _clear_services_cache()
    else:
        from Acquire.Client import Wallet as _Wallet
        _Wallet.remove_service(service)




__all__ = ["create_qrcode", "has_qrcode"]


def has_qrcode():
    """Return whether or not we support creating QR codes"""
    try:
        import qrcode as _qrcode
        return _qrcode is not None
    except:
        return False


def create_qrcode(uri):
    """Return a QR code for the passed URI"""
    try:
        import qrcode as _qrcode
    except:
        from Acquire.Client import QRCodeError
        raise QRCodeError("Cannot find the qrcode library needed to generate "
                          "QR codes. Please install and try again.")

    return _qrcode.make(uri)





__all__ = ["WeakPassphraseError", "KeyManipulationError",
           "SignatureVerificationError",
           "DecryptionError", "OTPError"]


class WeakPassphraseError(Exception):
    pass


class KeyManipulationError(Exception):
    pass


class SignatureVerificationError(Exception):
    pass


class DecryptionError(Exception):
    pass


class OTPError(Exception):
    pass




__all__ = ["UserDrives"]

_drives_root = "storage/drives"

_subdrives_root = "storage/subdrives"


class UserDrives:
    """This class holds all of the information about all of
       the drives that a specific user can access. This holds
       the relationships between those user drives, e.g.
       drives can be nested in other drives.
    """
    def __init__(self, authorisation=None, user_guid=None):
        """Construct either from a user-authorisation or specifying
           the user's GUID directly
        """
        if authorisation is not None:
            from Acquire.Identity import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError(
                    "You can only authorise UserDrives with a valid "
                    "Authorisation object")

            authorisation.verify(resource="UserDrives")

            self._is_authorised = True

            if user_guid is not None:
                if authorisation.user_guid() != user_guid:
                    raise PermissionError(
                        "Disagreement of user_guid: %s versus %s" %
                        (authorisation.user_guid(), user_guid))

            self._user_guid = authorisation.user_guid()
        else:
            self._user_guid = str(user_guid)
            self._is_authorised = False

    def is_null(self):
        """Return whether or not this is null"""
        return self._user_guid is None

    def list_drives(self, drive_uid=None):
        """Return a list of all of the top-level drives to which this user
           has access, or all of the sub-drives of the drive with
           passed 'drive_uid'
        """
        if self.is_null():
            return []

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import encoded_to_string as _encoded_to_string
        from Acquire.Storage import DriveMeta as _DriveMeta

        bucket = _get_service_account_bucket()

        if drive_uid is None:
            # look for the top-level drives
            names = _ObjectStore.get_all_object_names(
                        bucket, "%s/%s" % (_drives_root, self._user_guid))
        else:
            # look for the subdrives
            names = _ObjectStore.get_all_object_names(
                        bucket, "%s/%s/%s" %
                        (_subdrives_root, self._user_guid, drive_uid))

        drives = []
        for name in names:
            drive_name = _encoded_to_string(name.split("/")[-1])
            drives.append(_DriveMeta(name=drive_name, container=drive_uid))

        return drives

    def _get_subdrive(self, drive_uid, name, autocreate=True):
        """Return the DriveInfo for the Drive that the user has
           called 'name' in the drive with UID 'drive_uid'. If
           'autocreate' is True then this drive is automatically
           created if it does not exist.
        """
        if self.is_null():
            raise PermissionError(
                "You cannot get a DriveInfo from a null UserDrives")

        from Acquire.ObjectStore import string_to_filepath_parts \
            as _string_to_filepath_parts

        parts = _string_to_filepath_parts(name)

        if len(parts) != 1:
            raise ValueError(
                "The passed drive name '%s' is not valid!" % name)

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        from Acquire.ObjectStore import string_to_encoded as _string_to_encoded

        encoded_name = _string_to_encoded(name)

        bucket = _get_service_account_bucket()

        drive_key = "%s/%s/%s/%s" % (_subdrives_root, self._user_guid,
                                     drive_uid, encoded_name)

        try:
            drive_uid = _ObjectStore.get_string_object(
                                                bucket, drive_key)
        except:
            drive_uid = None

        if drive_uid is not None:
            from Acquire.Storage import DriveInfo as _DriveInfo
            drive = _DriveInfo(drive_uid=drive_uid, user_guid=self._user_guid,
                               is_authorised=self._is_authorised)
        else:
            drive = None

        if drive is None:
            if self._is_authorised and autocreate:
                # create a new UID for the drive and write this to the
                # object store
                from Acquire.ObjectStore import create_uuid as _create_uuid

                drive_uid = _create_uuid()

                drive_uid = _ObjectStore.set_ins_string_object(
                                            bucket, drive_key, drive_uid)

                from Acquire.Storage import DriveInfo as _DriveInfo
                drive = _DriveInfo(drive_uid=drive_uid,
                                   user_guid=self._user_guid,
                                   is_authorised=self._is_authorised)

        return drive

    def get_drive(self, name, autocreate=True):
        """Return the DriveMeta for the Drive that the user has
           called 'name'. If 'autocreate' is True then this
           drive is automatically created if it does not exist. Note
           that the '/' in the name will be interpreted as drive
           separators.
        """
        if self.is_null():
            raise PermissionError(
                "You cannot get a DriveInfo from a null UserDrives")

        # break the name into directory parts
        from Acquire.ObjectStore import string_to_filepath_parts \
            as _string_to_filepath_parts

        parts = _string_to_filepath_parts(name)

        # first get the root drive...
        root_name = parts[0]

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        from Acquire.ObjectStore import string_to_encoded as _string_to_encoded

        encoded_name = _string_to_encoded(root_name)
        drive_name = root_name

        bucket = _get_service_account_bucket()

        drive_key = "%s/%s/%s" % (_drives_root, self._user_guid,
                                  encoded_name)

        try:
            drive_uid = _ObjectStore.get_string_object(
                                                bucket, drive_key)
        except:
            drive_uid = None

        if drive_uid is not None:
            from Acquire.Storage import DriveInfo as _DriveInfo
            drive = _DriveInfo(drive_uid=drive_uid, user_guid=self._user_guid,
                               is_authorised=self._is_authorised)
        else:
            drive = None

        if drive is None:
            if self._is_authorised and autocreate:
                # create a new UID for the drive and write this to the
                # object store
                from Acquire.ObjectStore import create_uuid as _create_uuid

                drive_uid = _create_uuid()

                drive_uid = _ObjectStore.set_ins_string_object(
                                            bucket, drive_key, drive_uid)

                from Acquire.Storage import DriveInfo as _DriveInfo
                drive = _DriveInfo(drive_uid=drive_uid,
                                   user_guid=self._user_guid,
                                   is_authorised=self._is_authorised)

        if drive is None:
            from Acquire.Storage import MissingDriveError
            raise MissingDriveError(
                "There is no Drive called '%s' available" % name)

        container = []

        # now we have the drive, get the sub-drive in this drive...
        if len(parts) > 1:
            for subdrive in parts[1:]:
                container.append(drive.uid())
                drive_name = subdrive
                drive = self._get_subdrive(
                                    drive_uid=drive.uid(),
                                    name=drive_name,
                                    autocreate=autocreate)

                if drive is None:
                    from Acquire.Storage import MissingDriveError
                    raise MissingDriveError(
                        "There is no Drive called '%s' available" % name)

        from Acquire.Storage import DriveMeta as _DriveMeta

        return _DriveMeta(name=drive_name, uid=drive.uid(),
                          container=container,
                          aclrules=drive.aclrules())



"""
Acquire : (C) Christopher Woods 2018

System that allows users to log into a distributed serverless
function service to run functions. Data from functions is stored
in an intermediary Object Store. The complete system comprises
three services:

(1) Identity Service (Acquire.Identity)
    This manages user accounts and allows users to authenticate

(2) Accounting Service (Acquire.Accounting)
    This
"""

from Acquire.Stubs import lazy_import as _lazy_import

Access = _lazy_import.lazy_module("Acquire.Access")
Accounting = _lazy_import.lazy_module("Acquire.Accounting")
Crypto = _lazy_import.lazy_module("Acquire.Crypto")
Identity = _lazy_import.lazy_module("Acquire.Identity")
ObjectStore = _lazy_import.lazy_module("Acquire.ObjectStore")
Service = _lazy_import.lazy_module("Acquire.Service")
Client = _lazy_import.lazy_module("Acquire.Client")

__version__ = "0.0.7"

__all__ = ["Access", "Accounting", "Client", "Crypto",
           "Identity", "ObjectStore", "Service"]
